# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from hyperspy.signals import Signal1D


class PairDistributionFunction1D(Signal1D):
    """Signal class for Pair Distribution Function (PDF) radial profiles
    as a function of distance r.

    Parameters
    ----------
    *args
        See :class:`hyperspy.api.signals.Signal1D`.
    **kwargs
        See :class:`hyperspy.api.signals.Signal1D`
    """

    _signal_type = "pair_distribution_function"

    def normalise_signal(self, s_min=0, inplace=False, *args, **kwargs):
        """
        Normalises the Reduced PDF signal to having a maximum of 1.
        This is applied to each pdf separately in a multidimensional signal.

        Parameters
        ----------
        s_min : float
            The minimum scattering vector s to be considered. Values at lower s
            are ignored. This is to prevent the effect of large oscillations at
            r=0. If not set, the full signal is used.
        inplace : bool
            If True (default), this signal is overwritten. Otherwise, returns a
            new signal.
        *args:
            Arguments to be passed to :meth:`hyperspy.api.signals.BaseSignal.map`.
        **kwargs:
            Keyword arguments to be passed to :meth:`hyperspy.api.signals.BaseSignal.map`.

        """
        s_scale = self.axes_manager.signal_axes[0].scale
        index_min = int(s_min / s_scale)

        def normalise_pdf_signal_to_max(z, index_min):
            max_val = np.max(z[index_min:])
            return np.divide(z, max_val)

        return self.map(
            normalise_pdf_signal_to_max,
            index_min=index_min,
            inplace=inplace,
            *args,
            **kwargs
        )



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for vectors."""

# TODO: Delete this entire module

import numpy as np
from pyxem.utils._deprecated import deprecated


@deprecated(since="0.18.0", removal="1.0.0")
def normalize_virtual_images(im):
    """Normalizes image intensity by dividing by maximum value.

    Parameters
    ----------
    im : np.array()
        Array of image intensities

    Returns
    -------
    imn : np.array()
        Array of normalized image intensities

    """
    imn = im / im.max()
    return imn


@deprecated(
    since="0.18.0", removal="1.0.0", alternative="utils.vectors.get_vector_mesh"
)
def get_vectors_mesh(g1_norm, g2_norm, g_norm_max, angle=0.0, shear=0.0):
    """
    Calculate vectors coordinates of a mesh defined by a norm, a rotation and
    a shear component.

    Parameters
    ----------
    g1_norm, g2_norm : float
        The norm of the two vectors of the mesh.
    g_norm_max : float
        The maximum value for the norm of each vector.
    angle : float, optional
        The rotation of the mesh in degree.
    shear : float, optional
        The shear of the mesh. It must be in the interval [0, 1].
        The default is 0.0.

    Returns
    -------
    np.ndarray
        x and y coordinates of the vectors of the mesh

    """
    from pyxem.utils.vectors import get_vectors_mesh

    return get_vectors_mesh(g1_norm, g2_norm, g_norm_max, angle, shear)



# -*- coding: utf-8 -*-
# Copyright 2016-2022 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import pytest
import numpy as np
from sklearn.cluster import DBSCAN

from hyperspy.signals import Signal2D, BaseSignal, Signal1D
import hyperspy.api as hs

from pyxem.signals import DiffractionVectors2D, DiffractionVectors1D


class TestDiffractionVectors2D:
    def setup_method(self):
        vectors = np.reshape(np.repeat(np.arange(0, 100), 2), (50, 4))
        self.vector = DiffractionVectors2D(vectors)

    def test_setup(self):
        assert isinstance(self.vector, DiffractionVectors2D)
        assert self.vector.ragged == False

    def test_magnitudes(self):
        magnitudes = self.vector.get_magnitudes()
        assert magnitudes.ragged == False
        mags = np.linalg.norm(self.vector.data[:, [0, 1]], axis=1)
        np.testing.assert_array_almost_equal(mags, magnitudes.data)
        assert len(magnitudes.axes_manager.signal_axes) == 1

    def test_filter_magnitudes(self):
        magnitudes = self.vector.filter_magnitude(min_magnitude=10, max_magnitude=40)
        mags = np.linalg.norm(self.vector.data[:, [0, 1]], axis=1)
        num_in_range = np.sum((mags > 10) * (mags < 40))
        assert num_in_range == magnitudes.data.shape[0]


class TestSingleDiffractionVectors2D:
    def setup_method(self):
        self.data = np.array(
            [
                [0.063776, 0.011958],
                [-0.035874, 0.131538],
                [0.035874, -0.131538],
                [0.035874, 0.143496],
                [-0.035874, -0.13951],
                [-0.115594, 0.123566],
                [0.103636, -0.11958],
                [0.123566, 0.151468],
            ]
        )
        self.vector = DiffractionVectors2D(self.data)

    def test_filter_magnitude_single(self):
        filtered_vectors = self.vector.filter_magnitude(0.15, 1.0)
        ans = np.array(
            [[-0.115594, 0.123566], [0.103636, -0.11958], [0.123566, 0.151468]]
        )
        np.testing.assert_almost_equal(filtered_vectors.data, ans)

    def test_filter_detector_edge_single(self):
        self.vector.detector_shape = (260, 240)
        self.vector.column_scale = [0.001, 0.001]
        self.vector.column_offsets = [130 * 0.001, 120 * 0.001]
        filtered_vectors = self.vector.filter_detector_edge(exclude_width=10)
        ans = np.array([[0.063776, 0.011958]])
        np.testing.assert_almost_equal(filtered_vectors.data, ans)

    def test_cluster(self):
        clusterer = DBSCAN(eps=0.1, min_samples=2)
        clustered = self.vector.cluster(clusterer, min_vectors=3)
        assert clustered.data.shape[1] == 3
        assert clustered.data.shape[0] == 8
        assert clustered.axes_manager.signal_axes[0].size == 3
        assert clustered.ivec["cluster"].data.shape[0] == 8
        assert isinstance(clustered, DiffractionVectors2D)

    def test_slice(self):
        slic = self.vector.ivec[:, self.vector.ivec[1] > 0]
        assert slic.data.shape[0] == 5
        assert slic.data.shape[1] == 2


class TestVector2DSubclass:
    @pytest.fixture()
    def vectors(self):
        vectors = np.random.random((2, 2, 20, 2))
        v = DiffractionVectors2D(vectors)
        v.offsets = np.array([0, 0])
        v.scales = np.array([1, 1])
        v.column_names = ["x", "y"]
        return v

    def test_setup(self, vectors):
        assert isinstance(vectors, DiffractionVectors2D)
        assert isinstance(vectors, Signal2D)
        assert vectors.axes_manager.signal_shape == (2, 20)
        assert vectors.column_names == ["x", "y"]
        np.testing.assert_array_equal(vectors.scales, [1, 1])
        np.testing.assert_array_equal(vectors.offsets, [0, 0])

    @pytest.mark.parametrize("item", [0, "x"])
    def test_slice(self, vectors, item):
        sliced = vectors.ivec[item]
        assert isinstance(sliced, DiffractionVectors1D)
        assert isinstance(sliced, Signal1D)
        assert sliced.axes_manager.signal_shape == (20,)

    def test_flatten(self, vectors):
        flatten_diffraction_vectors = vectors.flatten_diffraction_vectors()
        assert isinstance(flatten_diffraction_vectors, DiffractionVectors2D)

    def test_flatten_twice(self, vectors):
        flatten_diffraction_vectors = vectors.flatten_diffraction_vectors()
        flatten_diffraction_vectors = (
            flatten_diffraction_vectors.flatten_diffraction_vectors()
        )
        assert isinstance(flatten_diffraction_vectors, DiffractionVectors2D)

    def test_to_markers(self, vectors):
        markers = vectors.to_markers()
        assert isinstance(markers, hs.plot.markers.Points)
        s = Signal2D(np.ones((2, 2, 10, 10)))
        s.add_marker(markers)

    def test_gt(self, vectors):
        gt_vectors = vectors > 1.1
        np.testing.assert_array_equal(False, gt_vectors.data)

    def test_gte(self, vectors):
        gt_vectors = vectors >= 1.1
        np.testing.assert_array_equal(False, gt_vectors.data)

    def test_lt(self, vectors):
        lt_vectors = vectors < 1.1
        np.testing.assert_array_equal(True, lt_vectors.data)

    def test_lte(self, vectors):
        lt_vectors = vectors <= 1.1
        np.testing.assert_array_equal(True, lt_vectors.data)

    def test_num_rows(self, vectors):
        assert vectors.num_rows == 20

    def test_from_peaks(self):
        with pytest.raises(NotImplementedError):
            DiffractionVectors2D.from_peaks(BaseSignal(np.ones((2, 2, 10, 10))))



import numpy as np
from orix.quaternion import Rotation
from diffsims.generators.simulation_generator import SimulationGenerator
from pyxem.signals import Diffraction2D
from pyxem.data import si_phase
from skimage.draw import ellipse, disk
from scipy.ndimage import gaussian_filter
import numpy.typing as npt


def create_diffraction_pattern(
    simulation: SimulationGenerator,
    shape: tuple = (512, 512),
    direct_beam_position: tuple = None,
    radius: int = 20,
    num_electrons: int = None,
    in_plane_angle: float = 0,
    calibration: float = 0.01,
    mirrored: bool = False,
    transformation_matrix: npt.NDArray = None,
):
    """
    Create a simulated (spot) diffraction pattern based on the provided simulation.

    Parameters
    ----------
    simulation: SimulationGenerator
        An instance of SimulationGenerator that contains the diffraction simulation data.
    shape: tuple
        The shape of the output diffraction pattern, e.g. (512, 512).
    direct_beam_position: tuple, optional
        The position of the direct beam in the diffraction pattern. If None, it defaults to the center of the shape.
    radius: int
        The radius of the disk used to simulate the diffraction spots in pixels.
    num_electrons: int, optional
        The number of electrons to simulate in the diffraction pattern. If None, no Poisson noise is applied.
    in_plane_angle: float
        The in-plane angle for rotating the diffraction pattern.
    calibration:
        The calibration factor for the diffraction pattern, in units of Angstroms per pixel.
    mirrored: bool
        If True, the diffraction pattern will be mirrored.
    transformation_matrix:
        A transformation matrix to apply to the diffraction pattern coordinates.
        If None, no transformation is applied.

    Returns
    -------
    numpy.ndarray
        A 2D array representing the simulated diffraction pattern.

    """
    if direct_beam_position is None:
        direct_beam_position = (shape[1] // 2, shape[0] // 2)
    transformed = simulation._get_transformed_coordinates(
        in_plane_angle,
        direct_beam_position,
        mirrored,
        units="pixel",
        calibration=calibration,
    )
    in_frame = (
        (transformed.data[:, 0] >= 0)
        & (transformed.data[:, 0] < shape[1])
        & (transformed.data[:, 1] >= 0)
        & (transformed.data[:, 1] < shape[0])
    )
    spot_coords = np.round(transformed.data[in_frame]).astype(int)
    if transformation_matrix is not None:
        direct_beam_position = direct_beam_position + (0,)
        spot_coords = (
            (spot_coords - direct_beam_position) @ transformation_matrix
        ) + direct_beam_position
    spot_intens = transformed.intensity[in_frame]
    pattern = np.zeros(shape)
    # checks that we have some spots
    if spot_intens.shape[0] == 0:
        return pattern
    else:
        for cord, inten in zip(spot_coords, spot_intens):
            rr, cc = disk(cord[:2], radius, shape=shape)
            pattern[rr, cc] = inten
    if num_electrons is not None:
        total = np.sum(spot_intens) * radius**2 * np.pi
        pattern = np.random.poisson((pattern / total) * num_electrons)
    return np.divide(pattern, np.max(pattern))


def simulated_strain(
    navigation_shape: tuple = (32, 32),
    signal_shape: tuple = (512, 512),
    disk_radius: int = 20,
    num_electrons: int = 1e5,
    strain_matrix: npt.NDArray = None,
    lazy: bool = False,
):
    """
    Create a simulated strain map from a simulated diffraction pattern and a strain matrix.

    Parameters
    ----------
    navigation_shape: tuple
        The shape of the navigation axes, e.g. (32, 32).
    signal_shape: tuple
        The shape of the signal axes, e.g. (512, 512).
    disk_radius: int
        The radius of the disk used to create the diffraction pattern.
    num_electrons:
        The number of electrons (per pixel) to simulate in the diffraction pattern.
    strain_matrix:
        A 3x3 matrix representing the strain to apply to the diffraction pattern.
        If None, a default strain matrix is used.
    lazy: bool
        If True, the returned signal will be lazy, otherwise it will be eager.
        Default is False.

    Returns
    -------
    Diffraction2D
        A simulated diffraction pattern with applied strain.
    """
    if strain_matrix is None:
        strain_matrix = np.array([[0.1, 0.05, 0], [0.15, 0.2, 0], [0, 0, 1]])
    p = si_phase()
    gen = SimulationGenerator()
    rotations = Rotation.from_euler(
        [
            [0, 0, 0],
        ],
        degrees=True,
    )
    sim = gen.calculate_diffraction2d(
        phase=p, rotation=rotations, reciprocal_radius=1.5, max_excitation_error=0.1
    )

    precip = np.zeros(navigation_shape, dtype=float)
    r = navigation_shape[0] // 2
    c = navigation_shape[1] // 2
    r_rad = int(np.round(r * 0.5))
    c_rad = int(np.round(c * 0.7))
    rr, cc = ellipse(r, c, r_radius=r_rad, c_radius=c_rad)
    precip[rr, cc] = 1

    gaussian_filter(precip, sigma=3, output=precip)

    data = np.empty(precip.shape + signal_shape)

    for ind in np.ndindex(precip.shape):
        t = np.eye(3) + strain_matrix * precip[ind]
        data[ind] = create_diffraction_pattern(
            sim,
            shape=signal_shape,
            radius=disk_radius,
            num_electrons=num_electrons,
            transformation_matrix=t,
        )

    strained = Diffraction2D(data)

    strained.axes_manager.signal_axes.set(
        name=("kx", "ky"), units=r"$\AA^{-1}$", scale=0.01
    )
    strained.axes_manager.navigation_axes.set(name=("x", "y"), units="nm", scale=1.0)

    strained.calibration.center = None

    if lazy:
        strained = strained.as_lazy()

    return strained



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da
import hyperspy.api as hs
from matplotlib import pyplot as plt
from numpy.random import default_rng
from skimage.draw import circle_perimeter_aa
import scipy

from pyxem.signals import (
    Diffraction1D,
    Diffraction2D,
    LazyDiffraction2D,
    PolarDiffraction2D,
    DiffractionVectors,
)
from pyxem.data.dummy_data import make_diffraction_test_data as mdtd


class TestComputeAndAsLazy2D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyDiffraction2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert isinstance(s, Diffraction2D)
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_4d_data_compute(self):
        dask_array = da.random.random((4, 4, 10, 15), chunks=(1, 1, 10, 15))
        s = LazyDiffraction2D(dask_array)
        s.compute()
        assert isinstance(s, Diffraction2D)
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = Diffraction2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyDiffraction2D)
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_4d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = Diffraction2D(data)
        s_lazy = s.as_lazy()
        assert isinstance(s_lazy, LazyDiffraction2D)
        assert data.shape == s_lazy.data.shape


class TestDecomposition:
    def test_decomposition_is_performed(self, diffraction_pattern):
        s = Diffraction2D(diffraction_pattern)
        s.decomposition()
        assert s.learning_results is not None

    def test_decomposition_class_assignment(self, diffraction_pattern):
        s = Diffraction2D(diffraction_pattern)
        s.decomposition()
        assert isinstance(s, Diffraction2D)


class TestAzimuthalIntegral1d:
    @pytest.fixture
    def ones(self):
        ones_diff = Diffraction2D(data=np.ones(shape=(10, 10)))
        ones_diff.axes_manager.signal_axes[0].scale = 0.1
        ones_diff.axes_manager.signal_axes[1].scale = 0.1
        ones_diff.axes_manager.signal_axes[0].name = "kx"
        ones_diff.axes_manager.signal_axes[1].name = "ky"
        ones_diff.unit = "2th_deg"
        return ones_diff

    def test_unit(self, ones):
        dif = Diffraction2D(data=[[1, 1], [1, 1]])
        assert dif.unit == "px"
        dif.unit = "!23"
        assert dif.unit == "px"

    def test_unit_set(self, ones):
        assert ones.unit == "2th_deg"

    def test_1d_azimuthal_integral_pyxem(self, ones):
        ones.calibration.center = None
        ones.calibration.scale = 0.2
        az = ones.get_azimuthal_integral1d(
            npt=10,
            inplace=True,
            radial_range=[0.0, 0.8],
            mean=True,
        )
        assert isinstance(ones, Diffraction1D)
        np.testing.assert_array_almost_equal(ones.data[0:8], np.ones(8))
        assert az is None

    def test_mask_mean(self, ones):
        ones.calibration.center = None
        ones.calibration.scale = 0.2
        mask = np.zeros(ones.data.shape, dtype=bool)
        mask[0:5, 0:5] = True
        ones.calibration.mask = mask
        az = ones.get_azimuthal_integral1d(
            npt=10,
            inplace=True,
            radial_range=[0.0, 0.8],
            mean=True,
        )
        assert isinstance(ones, Diffraction1D)
        np.testing.assert_array_almost_equal(ones.data[0:8], np.ones(8))
        assert az is None

    def test_1d_azimuthal_integral_inplace(self, ones):
        az = ones.get_azimuthal_integral1d(
            npt=10,
            inplace=True,
        )
        assert isinstance(ones, Diffraction1D)
        assert az is None

    def test_1d_azimuthal_integral_slicing(self, ones):
        ones.unit = "2th_rad"
        ones.calibration.center = None
        az = ones.get_azimuthal_integral1d(npt=10, radial_range=(0.0, 1.0), mean=True)
        np.testing.assert_array_almost_equal(az.data[0:7], np.ones(7))

    @pytest.mark.parametrize(
        "shape", [(20, 16), (3, 20, 16), (4, 3, 20, 16), (6, 4, 3, 20, 16)]
    )
    def test_lazy_input_lazy_output_different_shapes(self, shape):
        chunks = [5] * len(shape)
        s = LazyDiffraction2D(da.ones(shape, chunks=chunks))
        s.unit = "2th_deg"
        npt = 10
        s_a = s.get_azimuthal_integral1d(npt=npt)
        output_signal_shape = s.axes_manager.shape[:-2] + (npt,)
        output_data_shape = shape[:-2] + (npt,)
        assert s_a.axes_manager.shape == output_signal_shape
        assert s_a.data.shape == output_data_shape
        s_a.compute()
        assert s_a.axes_manager.shape == output_signal_shape
        assert s_a.data.shape == output_data_shape

    @pytest.mark.parametrize(
        "shape", [(20, 16), (3, 20, 16), (4, 3, 20, 16), (6, 4, 3, 20, 16)]
    )
    def test_non_lazy_input_lazy_output(self, shape):
        s = Diffraction2D(np.ones(shape))
        s.unit = "2th_deg"
        npt = 10
        s_a = s.get_azimuthal_integral1d(npt=npt, lazy_output=True)
        output_signal_shape = s.axes_manager.shape[:-2] + (npt,)
        output_data_shape = shape[:-2] + (npt,)
        assert s_a.axes_manager.shape == output_signal_shape
        assert s_a.data.shape == output_data_shape
        s_a.compute()
        assert s_a.axes_manager.shape == output_signal_shape
        assert s_a.data.shape == output_data_shape

    @pytest.mark.parametrize(
        "shape", [(20, 16), (3, 20, 16), (4, 3, 20, 16), (6, 4, 3, 20, 16)]
    )
    def test_lazy_input_non_lazy_output(self, shape):
        chunks = [5] * len(shape)
        s = LazyDiffraction2D(da.ones(shape, chunks=chunks))
        s.unit = "2th_deg"
        npt = 10
        s_a = s.get_azimuthal_integral1d(npt=npt, lazy_output=False)
        output_signal_shape = s.axes_manager.shape[:-2] + (npt,)
        output_data_shape = shape[:-2] + (npt,)
        assert s_a.axes_manager.shape == output_signal_shape
        assert s_a.data.shape == output_data_shape

    @pytest.mark.parametrize(
        "shape", [(20, 16), (3, 20, 16), (4, 3, 20, 16), (6, 4, 3, 20, 16)]
    )
    def test_lazy_input_lazy_result_inplace(self, shape):
        chunks = [5] * len(shape)
        s = LazyDiffraction2D(da.ones(shape, chunks=chunks))
        s.unit = "2th_deg"
        npt = 10
        output_signal_shape = s.axes_manager.shape[:-2] + (npt,)
        output_data_shape = shape[:-2] + (npt,)
        s.get_azimuthal_integral1d(npt=npt, inplace=True, lazy_output=True)
        assert s.axes_manager.shape == output_signal_shape
        assert s.data.shape == output_data_shape
        s.compute()
        assert s.axes_manager.shape == output_signal_shape
        assert s.data.shape == output_data_shape

    @pytest.mark.parametrize(
        "shape", [(20, 16), (3, 20, 16), (4, 3, 20, 16), (6, 4, 3, 20, 16)]
    )
    def test_lazy_input_non_lazy_result_inplace(self, shape):
        chunks = [5] * len(shape)
        s = LazyDiffraction2D(da.ones(shape, chunks=chunks))
        s.unit = "2th_deg"
        npt = 10
        output_signal_shape = s.axes_manager.shape[:-2] + (npt,)
        output_data_shape = shape[:-2] + (npt,)
        s.get_azimuthal_integral1d(npt=npt, inplace=True, lazy_output=False)
        assert s.axes_manager.shape == output_signal_shape
        assert s.data.shape == output_data_shape

    @pytest.mark.parametrize(
        "shape", [(20, 16), (3, 20, 16), (4, 3, 20, 16), (6, 4, 3, 20, 16)]
    )
    def test_non_lazy_input_lazy_result_inplace(self, shape):
        s = Diffraction2D(np.ones(shape))
        s.unit = "2th_deg"
        npt = 10
        output_signal_shape = s.axes_manager.shape[:-2] + (npt,)
        output_data_shape = shape[:-2] + (npt,)
        s.get_azimuthal_integral1d(npt=npt, inplace=True, lazy_output=True)
        assert s.axes_manager.shape == output_signal_shape
        assert s.data.shape == output_data_shape
        s.compute()
        assert s.axes_manager.shape == output_signal_shape
        assert s.data.shape == output_data_shape


class TestVariance:
    @pytest.fixture
    def ones(self):
        ones_diff = Diffraction2D(data=np.ones(shape=(10, 10, 10, 10)))
        ones_diff.calibration(scale=0.1, center=None)
        return ones_diff

    @pytest.fixture
    def ones_zeros(self):
        data = np.ones(shape=(10, 10, 10, 10))
        data[0:10:2, :, :, :] = 2
        ones_diff = Diffraction2D(data=data)
        ones_diff.calibration(scale=0.1, center=None)
        return ones_diff

    @pytest.fixture
    def bulls_eye_noisy(self):
        x, y = np.mgrid[-25:25, -25:25]
        data = (x**2 + y**2) ** 0.5
        data = np.tile(data, (5, 5, 1, 1))
        # Electron is equal to 1 count in image
        rng = default_rng(seed=1)
        data = rng.poisson(lam=data)
        ones_diff = Diffraction2D(data=data)
        ones_diff.calibration(scale=0.1, center=None)
        return ones_diff

    def test_FEM_Omega(self, ones, ones_zeros):
        ones_variance = ones.get_variance(npt=5, method="Omega")
        # assert ones_variance.axes_manager[0].units == "2th_deg"
        np.testing.assert_array_almost_equal(ones_variance.data, np.zeros(5), decimal=3)
        ones_zeros_variance = ones_zeros.get_variance(5, method="Omega")
        np.testing.assert_array_almost_equal(
            ones_zeros_variance.data, np.ones(5) * 0.1111, decimal=3
        )

    def test_FEM_Omega_poisson_noise(self, bulls_eye_noisy):
        bulls_eye_variance = bulls_eye_noisy.get_variance(25, method="Omega", dqe=1)
        # We exclude small radii
        np.testing.assert_array_almost_equal(
            bulls_eye_variance.data[5:], np.zeros(20), decimal=2
        )
        # Testing for non dqe=1
        bulls_eye_variance = (bulls_eye_noisy * 10).get_variance(
            25, method="Omega", dqe=10
        )
        # This fails at small radii and might still fail because it is random...
        np.testing.assert_array_almost_equal(
            bulls_eye_variance.data[6:], np.zeros(19), decimal=2
        )

    def test_FEM_r(self, ones, ones_zeros, bulls_eye_noisy):
        ones_variance = ones.get_variance(npt=5, method="r")
        # assert ones_variance.axes_manager[0].units == "2th_deg"
        np.testing.assert_array_almost_equal(ones_variance.data, np.zeros(5), decimal=3)
        ones_zeros_variance = ones_zeros.get_variance(5, method="r")
        np.testing.assert_array_almost_equal(
            ones_zeros_variance.data, np.zeros(5), decimal=3
        )
        bulls_eye_variance = bulls_eye_noisy.get_variance(25, method="r", dqe=1)
        # We exclude small radii
        np.testing.assert_array_almost_equal(
            bulls_eye_variance.data[5:], np.zeros(20), decimal=2
        )
        # Testing for non dqe=1
        bulls_eye_variance = (bulls_eye_noisy * 10).get_variance(25, method="r", dqe=10)
        # We exclude small radii
        np.testing.assert_array_almost_equal(
            bulls_eye_variance.data[6:], np.zeros(19), decimal=2
        )

    def test_FEM_r_spatial_kwarg(self, ones, ones_zeros, bulls_eye_noisy):
        v, fv = ones.get_variance(npt=5, method="r", spatial=True)

    @pytest.mark.parametrize("dqe_choice", [None, 0.3])
    def test_FEM_VImage(self, ones, dqe_choice):
        v = ones.get_variance(npt=5, method="VImage", dqe=dqe_choice)
        if dqe_choice is None:
            np.testing.assert_array_almost_equal(v.data, np.zeros(5), decimal=3)

    def test_FEM_re(self, ones, ones_zeros, bulls_eye_noisy):
        ones_variance = ones.get_variance(npt=5, method="re")
        np.testing.assert_array_almost_equal(ones_variance.data, np.zeros(5), decimal=3)
        ones_zeros_variance = ones_zeros.get_variance(5, method="re")
        np.testing.assert_array_almost_equal(
            ones_zeros_variance.data, np.ones(5) * 0.1111, decimal=3
        )
        bulls_eye_variance = bulls_eye_noisy.get_variance(25, method="re", dqe=1)
        # This fails at small radii and might still fail because it is random...
        np.testing.assert_array_almost_equal(
            bulls_eye_variance.data[5:], np.zeros(20), decimal=1
        )
        # Testing for non dqe=1
        bulls_eye_variance = (bulls_eye_noisy * 10).get_variance(
            25, method="re", dqe=10
        )
        # This fails at small radii and might still fail because it is random...
        np.testing.assert_array_almost_equal(
            bulls_eye_variance.data[6:], np.zeros(19), decimal=1
        )

    def test_not_existing_method(self, ones):
        with pytest.raises(ValueError):
            ones.get_variance(npt=5, method="magic")


class TestAzimuthalIntegral2d:
    @pytest.fixture
    def ones(self):
        ones_diff = Diffraction2D(data=np.ones(shape=(10, 10)))
        ones_diff.axes_manager.signal_axes[0].scale = 0.1
        ones_diff.axes_manager.signal_axes[1].scale = 0.1
        ones_diff.axes_manager.signal_axes[0].name = "kx"
        ones_diff.axes_manager.signal_axes[1].name = "ky"
        ones_diff.unit = "2th_deg"
        return ones_diff

    @pytest.fixture
    def arange(self):
        # signal looks as follows:
        # 0 1
        # 2 3
        arange_diff = Diffraction2D(data=np.arange(4).reshape(2, 2))
        arange_diff.axes_manager.signal_axes[0].scale = 1
        arange_diff.axes_manager.signal_axes[1].scale = 1
        arange_diff.axes_manager.signal_axes[0].name = "kx"
        arange_diff.axes_manager.signal_axes[1].name = "ky"
        arange_diff.unit = "2th_deg"
        return arange_diff

    @pytest.fixture
    def ring(self):
        ring_pattern = Diffraction2D(data=np.ones(shape=(100, 100)))
        rr, cc, val = circle_perimeter_aa(r=50, c=50, radius=30, shape=(100, 100))
        ring_pattern.data[rr, cc] = val * 100
        ring_pattern.axes_manager.signal_axes[0].scale = 0.1
        ring_pattern.axes_manager.signal_axes[1].scale = 0.1
        ring_pattern.axes_manager.signal_axes[0].name = "kx"
        ring_pattern.axes_manager.signal_axes[1].name = "ky"
        ring_pattern.unit = "k_nm^-1"
        return ring_pattern

    def test_internal_azimuthal_integration(self, ring):
        ring.calibration(scale=1)
        az = ring.get_azimuthal_integral2d(npt=40, npt_azim=100, radial_range=(0, 40))
        ring_sum = np.sum(az.data, axis=1)
        assert ring_sum.shape == (40,)

    @pytest.mark.parametrize(
        "corner",
        [
            (0, 0),
            (0, -1),
            (-1, 0),
            (-1, -1),
        ],
    )
    @pytest.mark.parametrize(
        "shape",
        [
            (10, 10),  # Even square
            (9, 9),  # Odd square
            (4, 6),  # Even
            (5, 9),  # Odd
            (5, 10),  # Odd and even
            (10, 5),  # Even and odd
        ],
    )
    def test_internal_azimuthal_integration_data_range(self, corner, shape):
        # Test that the edges of the cartesian data are included in the polar transform

        max_val = 10
        data = np.zeros(shape)
        data[corner] = max_val

        signal = Diffraction2D(data)

        # Reality check
        assert np.allclose(np.nanmax(signal.data), max_val)

        # Use mean=True to conserve values
        pol = signal.get_azimuthal_integral2d(npt=20, mean=True)

        assert np.allclose(np.nanmax(pol.data), max_val)

    # polar unwrapping `arange` should look like [1 0 2 3]
    # since data looks like:
    # 0 1
    # 2 3
    # and the data gets unwrapped from the center to the right
    @pytest.mark.parametrize(
        [
            "azimuthal_range",
            "expected_output",
        ],
        [
            [
                (0 * np.pi / 2, 1 * np.pi / 2),
                1,
            ],
            [
                (1 * np.pi / 2, 2 * np.pi / 2),
                0,
            ],
            [
                (2 * np.pi / 2, 3 * np.pi / 2),
                2,
            ],
            [
                (3 * np.pi / 2, 4 * np.pi / 2),
                3,
            ],
        ],
    )
    def test_azimuthal_integration_range(
        self, arange, azimuthal_range, expected_output
    ):
        arange.calibration.center = None  # set center
        quadrant = arange.get_azimuthal_integral2d(
            npt=10, npt_azim=10, azimuth_range=azimuthal_range, mean=True
        )
        assert np.allclose(quadrant.data[quadrant.data != 0], expected_output)


class TestVirtualImaging:
    # Tests that virtual imaging runs without failure

    def test_plot_integrated_intensity(self, diffraction_pattern):
        roi = hs.roi.CircleROI(3, 3, 5)
        plt.ion()  # to make plotting non-blocking
        diffraction_pattern.plot_integrated_intensity(roi)
        plt.close("all")

    def test_plot_line_profile(self, diffraction_pattern):
        diffraction_pattern.add_interactive_line_profile(linewidth=10)

    @pytest.mark.parametrize("has_nan", [True, False])
    def test_get_integrated_intensity(self, diffraction_pattern, has_nan):
        roi = hs.roi.CircleROI(3, 3, 5)
        if has_nan:
            diffraction_pattern.isig[:2] = np.nan
        vi = diffraction_pattern.get_integrated_intensity(roi)
        assert vi.axes_manager.signal_dimension == 2
        assert vi.axes_manager.navigation_dimension == 0
        np.testing.assert_allclose(vi.data, np.array([[6.0, 6.0], [8.0, 10.0]]))

    @pytest.mark.parametrize("out_signal_axes", [None, (0, 1), (1, 2), ("x", "y")])
    def test_get_integrated_intensity_stack(self, diffraction_pattern, out_signal_axes):
        s = hs.stack([diffraction_pattern] * 3)
        s.axes_manager.navigation_axes[0].name = "x"
        s.axes_manager.navigation_axes[1].name = "y"

        roi = hs.roi.CircleROI(3, 3, 5)
        vi = s.get_integrated_intensity(roi, out_signal_axes)
        assert vi.axes_manager.signal_dimension == 2
        assert vi.axes_manager.navigation_dimension == 1
        if out_signal_axes == (1, 2):
            assert vi.data.shape == (2, 3, 2)
            assert vi.axes_manager.navigation_size == 2
            assert vi.axes_manager.signal_shape == (2, 3)
        else:
            assert vi.data.shape == (3, 2, 2)
            assert vi.axes_manager.navigation_size == 3
            assert vi.axes_manager.signal_shape == (2, 2)

    def test_get_integrated_intensity_out_signal_axes(self, diffraction_pattern):
        s = hs.stack([diffraction_pattern] * 3)
        roi = hs.roi.CircleROI(3, 3, 5)
        vi = s.get_integrated_intensity(roi, out_signal_axes=(0, 1, 2))
        assert vi.axes_manager.signal_dimension == 3
        assert vi.axes_manager.navigation_dimension == 0
        assert vi.metadata.General.title == "Integrated intensity"
        assert (
            vi.metadata.Diffraction.integrated_range[:25] == "CircleROI(cx=3, cy=3, r=5"
        )

    def test_get_integrated_intensity_error(
        self, diffraction_pattern, out_signal_axes=(0, 1, 2)
    ):
        roi = hs.roi.CircleROI(3, 3, 5)
        with pytest.raises(ValueError):
            _ = diffraction_pattern.get_integrated_intensity(roi, out_signal_axes)

    def test_get_integrated_intensity_linescan(self, diffraction_pattern):
        s = diffraction_pattern.inav[0, :]
        s.metadata.General.title = ""
        roi = hs.roi.CircleROI(3, 3, 5)
        vi = s.get_integrated_intensity(roi)
        assert vi.data.shape == (2,)
        assert vi.axes_manager.signal_dimension == 1
        assert vi.axes_manager.navigation_dimension == 0
        assert (
            vi.metadata.Diffraction.integrated_range[:25] == "CircleROI(cx=3, cy=3, r=5"
        )


class TestAzimuthalIntegrator:
    # Tests the setting of a Azimutal Integrator:
    @pytest.fixture
    def ones(self):
        ones_diff = Diffraction2D(data=np.ones(shape=(10, 10)))
        ones_diff.axes_manager.signal_axes[0].scale = 0.1
        ones_diff.axes_manager.signal_axes[1].scale = 0.1
        ones_diff.axes_manager.signal_axes[0].name = "kx"
        ones_diff.axes_manager.signal_axes[1].name = "ky"
        ones_diff.unit = "2th_deg"
        return ones_diff

    @pytest.mark.xfail()
    def test_set_ai_fail(self, ones):
        # Not enough parameters fed into .set_ai()
        ones.unit = "k_nm^-1"
        ai = ones.set_ai()

    @pytest.mark.xfail()
    def test_return_ai_fail(self, ones):
        # .ai hasn't been set
        ai = ones.ai


class TestGetDirectBeamPosition:
    def setup_method(self):
        px, py, dx, dy = 15, 10, 20, 26
        x_pos_list = np.linspace(dx / 2 - 5, dx / 2 + 5, px, dtype=np.int16)
        y_pos_list = np.linspace(dy / 2 - 7, dy / 2 + 7, py, dtype=np.int16)
        data = np.zeros((py, px, dy, dx), dtype=np.uint16)
        for ix in range(data.shape[1]):
            for iy in range(data.shape[0]):
                x_pos, y_pos = x_pos_list[ix], y_pos_list[iy]
                data[iy, ix, y_pos, x_pos] = 10
        s = Diffraction2D(data)
        self.s = s
        self.dx, self.dy = dx, dy
        self.x_pos_list, self.y_pos_list = x_pos_list, y_pos_list

    @pytest.mark.parametrize(
        "sig_slice", (None, (2, 18, 2, 24), (2.0, 18.0, 2.0, 24.0))
    )
    @pytest.mark.parametrize(
        "method,kwargs",
        [
            ("cross_correlate", {"radius_start": 0, "radius_finish": 2}),
            (
                "blur",
                {
                    "sigma": 1,
                },
            ),
            (
                "interpolate",
                {
                    "sigma": 1,
                    "upsample_factor": 2,
                    "kind": "nearest",
                },
            ),
            ("center_of_mass", {"mask": (10, 13, 10)}),
        ],
    )
    def test_get_direct_beam(self, method, sig_slice, kwargs):
        s = self.s
        dx, dy = self.dx, self.dy
        s, x_pos_list, y_pos_list = self.s, self.x_pos_list, self.y_pos_list
        s_shift = s.get_direct_beam_position(
            method=method, signal_slice=sig_slice, **kwargs
        )
        if method == "cross_correlate":
            # shifted by half a pixel
            assert (-(x_pos_list - dx / 2) == s_shift.isig[0].data[0] + 0.5).all()
            assert (-(y_pos_list - dy / 2) == s_shift.isig[1].data[:, 0] + 0.5).all()
        else:
            np.testing.assert_array_almost_equal(
                -(x_pos_list - dx / 2), s_shift.isig[0].data[0]
            )
            np.testing.assert_array_almost_equal(
                -(y_pos_list - dy / 2), s_shift.isig[1].data[:, 0]
            )
        assert s.axes_manager.navigation_shape == s_shift.axes_manager.navigation_shape

    def test_fail_get_direct_beam(self):
        with pytest.raises(ValueError):
            s = self.s
            s_shift = s.get_direct_beam_position(
                method="blur",
                sigma=2,
                signal_slice=(5, 8, 5, 8),
                half_square_width=1,
            )

    def test_center_of_mass(self):
        dx, dy = self.dx, self.dy
        s, x_pos_list, y_pos_list = self.s, self.x_pos_list, self.y_pos_list
        s_shift = s.get_direct_beam_position(method="center_of_mass")
        assert s.axes_manager.navigation_shape == s_shift.axes_manager.navigation_shape
        assert (-(x_pos_list - dx / 2) == s_shift.isig[0].data[0]).all()
        assert (-(y_pos_list - dy / 2) == s_shift.isig[1].data[:, 0]).all()

    def test_lazy_result_none_non_lazy_signal(self):
        s = self.s
        s_shift = s.get_direct_beam_position(method="blur", sigma=1)
        assert not s_shift._lazy

    def test_lazy_result_none_lazy_signal(self):
        s = self.s.as_lazy()
        s_shift = s.get_direct_beam_position(method="blur", sigma=1)
        assert s_shift._lazy

    def test_lazy_result(self):
        s = self.s
        s_shift = s.get_direct_beam_position(method="blur", sigma=1, lazy_output=True)
        assert hasattr(s_shift.data, "compute")
        s_shift.compute()

    def test_lazy_input_non_lazy_result(self):
        s = LazyDiffraction2D(da.from_array(self.s.data))
        s_shift = s.get_direct_beam_position(method="blur", sigma=1, lazy_output=False)
        assert not hasattr(s_shift.data, "compute")
        assert not hasattr(s_shift, "compute")

    def test_lazy_input_lazy_result(self):
        s = LazyDiffraction2D(da.from_array(self.s.data))
        s_shift = s.get_direct_beam_position(method="blur", sigma=1, lazy_output=True)
        assert hasattr(s_shift.data, "compute")
        s_shift.compute()

    def test_non_uniform_chunks(self):
        s = LazyDiffraction2D(da.from_array(self.s.data, chunks=(8, 7, 10, 12)))
        s_shift = s.get_direct_beam_position(method="blur", sigma=1, lazy_output=True)
        shift_data_shape = s.data.shape[:-2] + (2,)
        assert s_shift.data.shape == shift_data_shape
        s_shift.compute()
        assert s_shift.data.shape == shift_data_shape


class TestDiffraction2DGetDirectBeamPositionCenterOfMass:
    def test_center_of_mass_0d(self):
        x_im, y_im, x, y = 7, 9, 2, 3
        array = np.zeros(shape=(y_im, x_im))
        array[y, x] = 1
        s = Diffraction2D(array)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()
        assert s_bs.axes_manager.navigation_shape == ()
        assert s_bs.axes_manager.signal_shape == (2,)

    def test_center_of_mass_1d(self):
        x_im, y_im, x, y = 7, 9, 2, 3
        array = np.zeros(shape=(5, y_im, x_im))
        array[:, y, x] = 1
        s = Diffraction2D(array)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()
        assert s_bs.axes_manager.navigation_shape == (5,)
        assert s_bs.axes_manager.signal_shape == (2,)

    def test_center_of_mass(self):
        x_im, y_im, x, y = 10, 10, 5, 7
        array = np.zeros(shape=(10, 10, y_im, x_im))
        array[:, :, y, x] = 1
        s = Diffraction2D(array)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()
        assert s_bs.axes_manager.navigation_shape == (10, 10)
        assert s_bs.axes_manager.signal_shape == (2,)

    def test_center_of_mass_random_position(self):
        x_im, y_im = 10, 10
        array = np.zeros(shape=(10, 10, y_im, x_im))
        x_array = np.random.randint(0, 10, size=(y_im, x_im))
        y_array = np.random.randint(0, 10, size=(y_im, x_im))
        for i in range(10):
            for j in range(10):
                array[i, j, y_array[i, j], x_array[i, j]] = 1
        s = Diffraction2D(array)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x_array, (y_im / 2) - y_array
        assert (s_bs.isig[0].data == x_new.astype(float)).all()
        assert (s_bs.isig[1].data == y_new.astype(float)).all()

    def test_center_of_mass_different_shapes(self):
        x_nav, y_nav, x_im, y_im = 10, 5, 8, 15
        array = np.zeros(shape=(y_nav, x_nav, y_im, x_im))
        x_array = np.random.randint(1, 7, size=(5, 10))
        y_array = np.random.randint(1, 14, size=(5, 10))
        for i in range(5):
            for j in range(10):
                array[i, j, y_array[i, j], x_array[i, j]] = 1
        s = Diffraction2D(array)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x_array, (y_im / 2) - y_array
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()
        assert s_bs.axes_manager.navigation_shape == (x_nav, y_nav)
        assert s_bs.axes_manager.signal_shape == (2,)

    def test_center_of_mass_different_shapes2(self):
        psX, psY = 11, 9
        s = mdtd.generate_4d_data(probe_size_x=psX, probe_size_y=psY, ring_x=None)
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        assert s_bs.axes_manager.shape == (psX, psY, 2)

    def test_different_shape_no_blur_no_downscale(self):
        y, x = np.mgrid[75:83:9j, 85:95:11j]
        x_im, y_im = 160, 140
        s = mdtd.generate_4d_data(
            probe_size_x=11,
            probe_size_y=9,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=40,
            disk_I=20,
            blur=False,
            blur_sigma=1,
            downscale=False,
        )
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()

    def test_different_shape_no_downscale(self):
        y, x = np.mgrid[75:83:9j, 85:95:11j]
        x_im, y_im = 160, 140
        s = mdtd.generate_4d_data(
            probe_size_x=11,
            probe_size_y=9,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=40,
            disk_I=20,
            blur=True,
            blur_sigma=1,
            downscale=False,
        )
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        np.testing.assert_allclose(s_bs.isig[0].data, x_new)
        np.testing.assert_allclose(s_bs.isig[1].data, y_new)

    def test_mask(self):
        y, x = np.mgrid[75:83:9j, 85:95:11j]
        x_im, y_im = 160, 140
        s = mdtd.generate_4d_data(
            probe_size_x=11,
            probe_size_y=9,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=40,
            disk_I=20,
            blur=False,
            blur_sigma=1,
            downscale=False,
        )
        s.data[:, :, 15, 10] = 1000000
        s_bs0 = s.get_direct_beam_position(method="center_of_mass")
        s_bs1 = s.get_direct_beam_position(method="center_of_mass", mask=(90, 79, 60))
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        assert not (s_bs0.isig[0].data == x_new).all()
        assert not (s_bs0.isig[1].data == y_new).all()
        assert (s_bs1.isig[0].data == x_new).all()
        assert (s_bs1.isig[1].data == y_new).all()

    def test_mask_2(self):
        x, y = 60, 50
        x_im, y_im = 120, 100
        s = mdtd.generate_4d_data(
            probe_size_x=5,
            probe_size_y=5,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=20,
            disk_I=20,
            blur=False,
            downscale=False,
        )
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y

        # Add one large value
        s.data[:, :, 50, 30] = 200000  # Large value to the left of the disk

        # Center of mass should not be in center of the disk, due to the
        # large value.
        s_bs1 = s.get_direct_beam_position(method="center_of_mass")
        assert not (s_bs1.isig[0].data == x_new).all()
        assert (s_bs1.isig[1].data == y_new).all()

        # Here, the large value is masked
        s_bs2 = s.get_direct_beam_position(method="center_of_mass", mask=(60, 50, 25))
        assert (s_bs2.isig[0].data == x_new).all()
        assert (s_bs2.isig[1].data == y_new).all()

        # Here, the large value is right inside the edge of the mask
        s_bs3 = s.get_direct_beam_position(method="center_of_mass", mask=(60, 50, 31))
        assert not (s_bs3.isig[0].data == x_new).all()
        assert (s_bs3.isig[1].data == y_new).all()

        # Here, the large value is right inside the edge of the mask
        s_bs4 = s.get_direct_beam_position(method="center_of_mass", mask=(59, 50, 30))
        assert not (s_bs4.isig[0].data == x_new).all()
        assert (s_bs4.isig[1].data == y_new).all()

        s.data[:, :, 50, 30] = 0
        s.data[:, :, 80, 60] = 200000  # Large value under the disk

        # The large value is masked
        s_bs5 = s.get_direct_beam_position(method="center_of_mass", mask=(60, 50, 25))
        assert (s_bs5.isig[0].data == x_new).all()
        assert (s_bs5.isig[1].data == y_new).all()

        # The large value just not masked
        s_bs6 = s.get_direct_beam_position(method="center_of_mass", mask=(60, 50, 31))
        assert (s_bs6.isig[0].data == x_new).all()
        assert not (s_bs6.isig[1].data == y_new).all()

        # The large value just not masked
        s_bs7 = s.get_direct_beam_position(method="center_of_mass", mask=(60, 55, 25))
        assert (s_bs7.isig[0].data == x_new).all()
        assert not (s_bs7.isig[1].data == y_new).all()

    def test_threshold(self):
        x_im, y_im = 120, 100
        x, y = 60, 50
        s = mdtd.generate_4d_data(
            probe_size_x=4,
            probe_size_y=3,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=20,
            disk_I=20,
            blur=False,
            blur_sigma=1,
            downscale=False,
        )
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        s.data[:, :, 0:30, 0:30] = 5

        # The extra values are ignored due to thresholding
        s_bs0 = s.get_direct_beam_position(method="center_of_mass", threshold=2)
        assert (s_bs0.isig[0].data == x_new).all()
        assert (s_bs0.isig[1].data == y_new).all()

        # The extra values are not ignored
        s_bs1 = s.get_direct_beam_position(method="center_of_mass", threshold=1)
        assert not (s_bs1.isig[0].data == x_new).all()
        assert not (s_bs1.isig[1].data == y_new).all()

        # The extra values are not ignored
        s_bs2 = s.get_direct_beam_position(method="center_of_mass")
        assert not (s_bs2.isig[0].data == x_new).all()
        assert not (s_bs2.isig[1].data == y_new).all()

    def test_threshold_and_mask(self):
        x_im, y_im = 120, 100
        x, y = 60, 50
        s = mdtd.generate_4d_data(
            probe_size_x=4,
            probe_size_y=3,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=20,
            disk_I=20,
            blur=False,
            blur_sigma=1,
            downscale=False,
        )
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        s.data[:, :, 0:30, 0:30] = 5
        s.data[:, :, 1, -2] = 60

        # The extra values are ignored due to thresholding and mask
        s_bs0 = s.get_direct_beam_position(
            method="center_of_mass", threshold=3, mask=(60, 50, 50)
        )
        assert (s_bs0.isig[0].data == x_new).all()
        assert (s_bs0.isig[1].data == y_new).all()

        # The extra values are not ignored
        s_bs1 = s.get_direct_beam_position(method="center_of_mass", mask=(60, 50, 50))
        assert not (s_bs1.isig[0].data == x_new).all()
        assert not (s_bs1.isig[1].data == y_new).all()

        # The extra values are not ignored
        s_bs2 = s.get_direct_beam_position(method="center_of_mass", threshold=3)
        assert not (s_bs2.isig[0].data == x_new).all()
        assert not (s_bs2.isig[1].data == y_new).all()

        # The extra values are not ignored
        s_bs3 = s.get_direct_beam_position(method="center_of_mass")
        assert not (s_bs3.isig[0].data == x_new).all()
        assert not (s_bs3.isig[1].data == y_new).all()

    def test_1d_signal(self):
        x_im, y_im = 120, 100
        x = np.arange(45, 45 + 9).reshape((1, 9))
        y = np.arange(55, 55 + 9).reshape((1, 9))
        s = mdtd.generate_4d_data(
            probe_size_x=9,
            probe_size_y=1,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=20,
            disk_I=20,
            blur=False,
            blur_sigma=1,
            downscale=False,
        )
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        s_bs = s.inav[:, 0].get_direct_beam_position(method="center_of_mass")
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()

    def test_0d_signal(self):
        x_im, y_im = 120, 100
        x, y = 40, 51
        s = mdtd.generate_4d_data(
            probe_size_x=1,
            probe_size_y=1,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=20,
            disk_I=20,
            blur=False,
            blur_sigma=1,
            downscale=False,
        )
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        s_bs = s.inav[0, 0].get_direct_beam_position(method="center_of_mass")
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()

    def test_lazy(self):
        x_im, y_im = 160, 140
        y, x = np.mgrid[75:83:9j, 85:95:11j]
        s = mdtd.generate_4d_data(
            probe_size_x=11,
            probe_size_y=9,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=40,
            disk_I=20,
            blur=True,
            blur_sigma=1,
            downscale=False,
        )
        x_new, y_new = (x_im / 2) - x, (y_im / 2) - y
        s_lazy = LazyDiffraction2D(da.from_array(s.data, chunks=(1, 1, 140, 160)))
        s_lazy_bs = s_lazy.get_direct_beam_position(method="center_of_mass")
        np.testing.assert_allclose(s_lazy_bs.isig[0].data, x_new)
        np.testing.assert_allclose(s_lazy_bs.isig[1].data, y_new)

        s_lazy_1d = s_lazy.inav[0]
        s_lazy_1d_bs = s_lazy_1d.get_direct_beam_position(method="center_of_mass")
        np.testing.assert_allclose(s_lazy_1d_bs.isig[0].data, x_new[:, 0])
        np.testing.assert_allclose(s_lazy_1d_bs.isig[1].data, y_new[:, 0])

        s_lazy_0d = s_lazy.inav[0, 0]
        s_lazy_0d_bs = s_lazy_0d.get_direct_beam_position(method="center_of_mass")
        np.testing.assert_allclose(s_lazy_0d_bs.isig[0].data, x_new[0, 0])
        np.testing.assert_allclose(s_lazy_0d_bs.isig[1].data, y_new[0, 0])

    def test_compare_lazy_and_nonlazy(self):
        x_im, y_im = 160, 140
        y, x = np.mgrid[75:83:9j, 85:95:11j]
        s = mdtd.generate_4d_data(
            probe_size_x=11,
            probe_size_y=9,
            ring_x=None,
            image_size_x=x_im,
            image_size_y=y_im,
            disk_x=x,
            disk_y=y,
            disk_r=40,
            disk_I=20,
            blur=True,
            blur_sigma=1,
            downscale=False,
        )
        s_lazy = LazyDiffraction2D(da.from_array(s.data, chunks=(1, 1, 140, 160)))
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        s_lazy_bs = s_lazy.get_direct_beam_position(method="center_of_mass")
        np.testing.assert_equal(s_bs.data, s_lazy_bs.data)

        bs_nav_extent = s_bs.axes_manager.navigation_extent
        lazy_bs_nav_extent = s_lazy_bs.axes_manager.navigation_extent
        assert bs_nav_extent == lazy_bs_nav_extent

        bs_sig_extent = s_bs.axes_manager.signal_extent
        lazy_bs_sig_extent = s_lazy_bs.axes_manager.signal_extent
        assert bs_sig_extent == lazy_bs_sig_extent

    def test_lazy_result(self):
        data = da.ones((10, 10, 20, 20), chunks=(10, 10, 10, 10))
        s_lazy = LazyDiffraction2D(data)
        s_lazy_bs = s_lazy.get_direct_beam_position(
            method="center_of_mass", lazy_output=True
        )
        assert s_lazy_bs._lazy
        assert s_lazy_bs.axes_manager.navigation_shape == (10, 10)

        s_lazy_1d = s_lazy.inav[0]
        s_lazy_1d_bs = s_lazy_1d.get_direct_beam_position(
            method="center_of_mass", lazy_output=True
        )
        assert s_lazy_1d_bs._lazy
        assert s_lazy_1d_bs.axes_manager.navigation_shape == (10,)

        s_lazy_0d = s_lazy.inav[0, 0]
        s_lazy_0d_bs = s_lazy_0d.get_direct_beam_position(
            method="center_of_mass", lazy_output=True
        )
        assert s_lazy_0d_bs._lazy
        assert s_lazy_0d_bs.axes_manager.navigation_shape == ()

    def test_center_of_mass_inplace(self):
        with pytest.raises(ValueError):
            d = Diffraction2D(np.zeros((10, 10, 20, 20)))
            d.get_direct_beam_position(method="center_of_mass", inplace=True)


class TestCenterDirectBeam:
    def setup_method(self):
        data = np.zeros((8, 6, 20, 16), dtype=np.int16)
        x_pos_list = np.random.randint(8 - 2, 8 + 2, 6, dtype=np.int16)
        x_pos_list[x_pos_list == 8] = 9
        y_pos_list = np.random.randint(10 - 2, 10 + 2, 8, dtype=np.int16)
        for ix in range(len(x_pos_list)):
            for iy in range(len(y_pos_list)):
                data[iy, ix, y_pos_list[iy], x_pos_list[ix]] = 9
        s = Diffraction2D(data)
        s.axes_manager[0].scale = 0.5
        s.axes_manager[1].scale = 0.6
        s.axes_manager[2].scale = 3
        s.axes_manager[3].scale = 4
        s_lazy = s.as_lazy()
        self.s = s
        self.s_lazy = s_lazy
        self.x_pos_list = x_pos_list
        self.y_pos_list = y_pos_list

    def test_non_lazy(self):
        s = self.s
        s.center_direct_beam(method="blur", sigma=1)
        assert s._lazy is False
        assert (s.data[:, :, 10, 8] == 9).all()
        # Make sure only the pixel we expect to change, has actually changed
        s.data[:, :, 10, 8] = 0
        assert not s.data.any()

    def test_non_lazy_lazy_result(self):
        s = self.s
        s.center_direct_beam(method="blur", sigma=1, lazy_output=True, inplace=True)
        assert s._lazy is True
        s.compute()
        assert (s.data[:, :, 10, 8] == 9).all()
        s.data[:, :, 10, 8] = 0
        assert not s.data.any()

    def test_lazy(self):
        s_lazy = self.s_lazy
        s_lazy.center_direct_beam(method="blur", sigma=1)
        assert s_lazy._lazy is True
        s_lazy.compute()
        assert (s_lazy.data[:, :, 10, 8] == 9).all()
        s_lazy.data[:, :, 10, 8] = 0
        assert not s_lazy.data.any()

    def test_lazy_not_lazy_result(self):
        s_lazy = self.s_lazy
        s_lazy.center_direct_beam(method="blur", sigma=1, lazy_output=False)
        assert s_lazy._lazy is False
        assert (s_lazy.data[:, :, 10, 8] == 9).all()
        s_lazy.data[:, :, 10, 8] = 0
        assert not s_lazy.data.any()

    def test_non_uniform_chunks(self):
        s_lazy = self.s_lazy
        s_lazy.data = s_lazy.data.rechunk((5, 4, 12, 14))
        s_lazy_shape = s_lazy.axes_manager.shape
        data_lazy_shape = s_lazy.data.shape
        s_lazy.center_direct_beam(method="blur", sigma=1, lazy_output=True)
        assert s_lazy.axes_manager.shape == s_lazy_shape
        assert s_lazy.data.shape == data_lazy_shape
        s_lazy.compute()
        assert s_lazy.axes_manager.shape == s_lazy_shape
        assert s_lazy.data.shape == data_lazy_shape

    def test_return_shifts_non_lazy(self):
        s = self.s
        s_shifts = s.center_direct_beam(method="blur", sigma=1, return_shifts=True)
        assert s_shifts._lazy is False
        nav_dim = s.axes_manager.navigation_dimension
        assert nav_dim == s_shifts.axes_manager.navigation_dimension
        x_pos_list, y_pos_list = self.x_pos_list, self.y_pos_list
        assert ((8 - x_pos_list) == s_shifts.isig[0].data[0]).all()
        assert ((10 - y_pos_list) == s_shifts.isig[1].data[:, 0]).all()

    def test_return_shifts_lazy(self):
        s_lazy = self.s_lazy
        s_shifts = s_lazy.center_direct_beam(method="blur", sigma=1, return_shifts=True)
        assert s_shifts._lazy is True
        s_shifts.compute()
        x_pos_list, y_pos_list = self.x_pos_list, self.y_pos_list
        assert ((8 - x_pos_list) == s_shifts.isig[0].data[0]).all()
        assert ((10 - y_pos_list) == s_shifts.isig[1].data[:, 0]).all()

    def test_return_axes_manager(self):
        s = self.s
        s.center_direct_beam(method="blur", sigma=1)
        assert s.axes_manager[0].scale == 0.5
        assert s.axes_manager[1].scale == 0.6
        assert s.axes_manager[2].scale == 3
        assert s.axes_manager[3].scale == 4

    def test_shifts_input(self):
        s = self.s
        s_shifts = s.get_direct_beam_position(method="blur", sigma=1, lazy_output=False)
        s.center_direct_beam(shifts=s_shifts)
        assert (s.data[:, :, 10, 8] == 9).all()
        s.data[:, :, 10, 8] = 0
        assert not s.data.any()

    def test_shifts_input_lazy(self):
        s = self.s
        s_shifts = s.get_direct_beam_position(method="blur", sigma=1, lazy_output=True)
        s.center_direct_beam(shifts=s_shifts)
        assert (s.data[:, :, 10, 8] == 9).all()
        s.data[:, :, 10, 8] = 0
        assert not s.data.any()

    def test_subpixel(self):
        s = self.s
        s_shifts = s.get_direct_beam_position(method="blur", sigma=1)
        s_shifts += 0.5
        s.change_dtype("float32")
        s.center_direct_beam(shifts=s_shifts, subpixel=True)
        assert (s.data[:, :, 10:12, 8:10] == 9 / 4).all()
        s.data[:, :, 10:12, 8:10] = 0.0
        assert not s.data.any()

    def test_not_subpixel(self):
        s = self.s
        s_shifts = s.get_direct_beam_position(method="blur", sigma=1)
        s_shifts += 0.3
        s.change_dtype("float32")
        s.center_direct_beam(shifts=s_shifts, subpixel=False)
        assert (s.data[:, :, 10, 8] == 9).all()
        s.data[:, :, 10, 8] = 0.0
        assert not s.data.any()

    @pytest.mark.parametrize(
        "shape", [(20, 20), (10, 20, 20), (8, 10, 20, 20), (6, 8, 10, 20, 20)]
    )
    def test_different_dimensions(self, shape):
        s = Diffraction2D(np.random.randint(0, 256, size=shape))
        s.center_direct_beam(method="blur", sigma=1)
        assert s.data.shape == shape

    def test_half_square_width(self):
        # Generating a larger dataset to check that half_square_width
        # works properly with the automatic chunking in dask_tools._get_dask_array
        s = Diffraction2D(np.zeros((10, 10, 200, 200)))
        s.axes_manager.signal_axes[0].offset = 100
        s.axes_manager.signal_axes[1].offset = 50
        x_pos_list = np.random.randint(100 - 2, 100 + 2, 10, dtype=np.int16)
        y_pos_list = np.random.randint(100 - 2, 100 + 2, 10, dtype=np.int16)
        for ix in range(len(x_pos_list)):
            for iy in range(len(y_pos_list)):
                s.data[iy, ix, y_pos_list[iy], x_pos_list[ix]] = 9
        s.data[:, :, 1, -1] = 1000

        s1 = s.deepcopy()
        s.center_direct_beam(method="blur", sigma=1)
        assert (s.data[:, :, 100, 100] == 1000).all()
        s1.center_direct_beam(method="blur", sigma=1, half_square_width=5)
        assert (s1.data[:, :, 100, 100] == 9).all()

    def test_align_kwargs(self):
        s = self.s
        s.data += 1
        s1 = s.deepcopy()
        s.center_direct_beam(method="blur", sigma=1)
        assert (s.data == 0).any()
        s1.center_direct_beam(method="blur", sigma=1, align_kwargs={"mode": "wrap"})
        assert not (s1.data == 0).any()

    def test_method_interpolate(self):
        s = self.s
        s.center_direct_beam(method="interpolate", sigma=1, upsample_factor=10, kind=1)

    def test_method_cross_correlate(self):
        s = self.s
        s.center_direct_beam(method="cross_correlate", radius_start=0, radius_finish=2)

    def test_method_center_of_mass(self):
        s = self.s
        s.center_direct_beam(method="center_of_mass")

    def test_parameter_both_method_and_shifts(self):
        s = self.s
        with pytest.raises(ValueError):
            s.center_direct_beam(method="blur", sigma=1, shifts=np.ones((8, 6, 2)))

    def test_parameter_neither_method_and_shifts(self):
        s = self.s
        with pytest.raises(ValueError):
            s.center_direct_beam()


class TestFindVectors:
    def setup_method(self):
        data = np.zeros((8, 6, 20, 16), dtype=np.int16)
        x_pos_list = np.random.randint(8 - 2, 8 + 2, 6, dtype=np.int16)
        x_pos_list[x_pos_list == 8] = 9
        y_pos_list = np.random.randint(10 - 2, 10 + 2, 8, dtype=np.int16)
        for ix in range(len(x_pos_list)):
            for iy in range(len(y_pos_list)):
                data[iy, ix, y_pos_list[iy], x_pos_list[ix]] = 9
        s = Diffraction2D(data)
        s.axes_manager[0].scale = 0.5
        s.axes_manager[1].scale = 0.6
        s.axes_manager[2].scale = 3
        s.axes_manager[3].scale = 4
        s_lazy = s.as_lazy()
        self.s = s
        self.s_lazy = s_lazy
        self.x_pos_list = x_pos_list
        self.y_pos_list = y_pos_list

    def test_find_vectors(self):
        s = self.s
        vectors = s.get_diffraction_vectors()
        assert isinstance(vectors, DiffractionVectors)
        assert vectors.column_names == ["ky", "kx", "intensity"]
        assert vectors.units == ["px", "px", "a.u."]


class TestSubtractingDiffractionBackground:
    method1 = ["difference of gaussians", "median kernel", "radial median", "h-dome"]

    def setup_method(self):
        self.data = np.random.rand(3, 2, 20, 15)
        self.data[:, :, 10:12, 7:9] = 100
        self.dp = Diffraction2D(self.data)

    @pytest.mark.parametrize("methods", method1)
    def test_subtract_backgrounds(self, methods):
        # test that background is mostly removed
        if methods == "h-dome":
            kwargs = {"h": 0.25}
        else:
            kwargs = {}
        subtracted = self.dp.subtract_diffraction_background(method=methods, **kwargs)
        assert isinstance(subtracted, Diffraction2D)
        assert subtracted.data.shape == self.data.shape

    def test_exception_not_implemented_method(self):
        s = Diffraction2D(np.zeros((2, 2, 10, 10)))
        with pytest.raises(NotImplementedError):
            s.subtract_diffraction_background(method="magic")


class TestFindHotPixels:
    @pytest.fixture()
    def hot_pixel_data(self):
        """Values are 50, except [21, 11] and [5, 38]
        being 50000 (to represent a "hot pixel").
        """
        data = np.ones((2, 2, 40, 50)) * 50
        data[:, :, 21, 11] = 50000
        data[:, :, 5, 38] = 50000
        dask_array = da.from_array(data, chunks=(1, 1, 40, 50))
        return LazyDiffraction2D(dask_array)

    @pytest.mark.parametrize("lazy", (True, False))
    def test_find_hot_pixels(self, hot_pixel_data, lazy):
        if not lazy:
            hot_pixel_data.compute()
            assert not hot_pixel_data._lazy
        s_hot_pixels = hot_pixel_data.find_hot_pixels()
        assert s_hot_pixels._lazy == lazy
        assert s_hot_pixels.data[0, 0, 21, 11]
        assert s_hot_pixels.data[0, 0, 5, 38]
        s_hot_pixels.data[:, :, 21, 11] = False
        s_hot_pixels.data[:, :, 5, 38] = False
        assert not s_hot_pixels.data.any()

    def test_threshold_multiplier(self, hot_pixel_data):
        s_hot_pixels = hot_pixel_data.find_hot_pixels(threshold_multiplier=1000000)
        assert not s_hot_pixels.data.any()

    def test_mask_array(self, hot_pixel_data):
        mask_array = Diffraction2D(np.ones_like(hot_pixel_data.data, dtype=bool))
        s_hot_pixels = hot_pixel_data.find_hot_pixels(mask=mask_array)
        assert not s_hot_pixels.data.any()


class TestFindDeadPixels:
    @pytest.fixture()
    def dead_pixel_data_2d(self):
        """Values are 50, except [14, 42] and [2, 12]
        being 0 (to represent a "dead pixel").
        """
        data = np.ones((40, 50)) * 50
        data[14, 42] = 0
        data[2, 12] = 0
        dask_array = da.from_array(data, chunks=(5, 5))
        return LazyDiffraction2D(dask_array)

    @pytest.mark.parametrize("lazy_case", (True, False))
    def test_2d(self, dead_pixel_data_2d, lazy_case):
        if not lazy_case:
            dead_pixel_data_2d.compute()
        s_dead_pixels = dead_pixel_data_2d.find_dead_pixels()
        assert s_dead_pixels.data.shape == dead_pixel_data_2d.data.shape
        assert s_dead_pixels.data[14, 42]
        assert s_dead_pixels.data[2, 12]
        s_dead_pixels.data[14, 42] = False
        s_dead_pixels.data[2, 12] = False
        assert not s_dead_pixels.data.any()

    def test_3d(self):
        data = np.ones((5, 40, 50)) * 50
        data[:, 14, 42] = 0
        data[:, 2, 12] = 0
        dask_array = da.from_array(data, chunks=(5, 5, 5))
        s = LazyDiffraction2D(dask_array)
        s_dead_pixels = s.find_dead_pixels()
        assert s_dead_pixels.data.shape == s.data.shape[-2:]

    def test_4d(self):
        data = np.ones((10, 5, 40, 50)) * 50
        data[:, :, 14, 42] = 0
        data[:, :, 2, 12] = 0
        dask_array = da.from_array(data, chunks=(5, 5, 5, 5))
        s = LazyDiffraction2D(dask_array)
        s_dead_pixels = s.find_dead_pixels()
        assert s_dead_pixels.data.shape == s.data.shape[-2:]

    def test_dead_pixel_value(self, dead_pixel_data_2d):
        s_dead_pixels = dead_pixel_data_2d.find_dead_pixels(dead_pixel_value=-10)
        assert not s_dead_pixels.data.any()

    def test_mask_array(self, dead_pixel_data_2d):
        mask_array = np.ones_like(dead_pixel_data_2d.data, dtype=bool)
        s_dead_pixels = dead_pixel_data_2d.find_dead_pixels(mask=mask_array)
        assert not s_dead_pixels.data.any()


class TestCorrectBadPixel:
    @pytest.fixture()
    def data(self):
        data = np.ones((2, 2, 100, 90))
        data[:, :, 9, 81] = 50000
        data[:, :, 41, 21] = 0
        return data

    @pytest.fixture()
    def bad_pixels(self):
        return np.asarray([[41, 21], [9, 81]])

    @pytest.mark.xfail(reason="This array shape is not currently supported")
    def test_lazy(self, data, bad_pixels):
        s_lazy = Diffraction2D(data).as_lazy()
        s_lazy.correct_bad_pixels(bad_pixels, lazy_result=True)

    @pytest.mark.parametrize("lazy_input", (True, False))
    @pytest.mark.parametrize("lazy_result", (True, False))
    def test_lazy_with_bad_pixel_finders(self, data, lazy_input, lazy_result):
        s = Diffraction2D(data).as_lazy()
        if not lazy_input:
            s.compute()

        hot = s.find_hot_pixels(lazy_output=True)
        dead = s.find_dead_pixels()
        bad = hot + dead

        s = s.correct_bad_pixels(bad, lazy_output=lazy_result, inplace=False)
        assert s._lazy == lazy_result
        if lazy_result:
            s.compute()
        assert np.isclose(s.data[0, 0, 9, 81], 1)
        assert np.isclose(s.data[0, 0, 41, 21], 1)


class TestMakeProbeNavigation:
    def test_fast(self):
        s = Diffraction2D(np.ones((6, 5, 12, 10)))
        s.make_probe_navigation(method="fast")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape
        assert np.all(s_nav.data == 1)

    def test_slow(self):
        s = Diffraction2D(np.ones((5, 5, 12, 10)))
        s.make_probe_navigation(method="slow")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape
        assert np.all(s_nav.data == 120)

    def test_fast_lazy(self):
        s = LazyDiffraction2D(da.ones((6, 4, 60, 50), chunks=(2, 2, 10, 10)))
        s.make_probe_navigation(method="fast")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape
        assert np.all(s_nav.data == 100)

    def test_slow_lazy(self):
        s = LazyDiffraction2D(da.ones((6, 4, 60, 50), chunks=(2, 2, 10, 10)))
        s.make_probe_navigation(method="slow")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape
        assert np.all(s_nav.data == 3000)

    def test_very_asymmetric_size(self):
        s = Diffraction2D(np.ones((50, 2, 120, 10)))
        s.make_probe_navigation(method="fast")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape

    def test_very_asymmetric_size_lazy(self):
        s = LazyDiffraction2D(da.ones((50, 2, 120, 10), chunks=(2, 2, 5, 5)))
        s.make_probe_navigation(method="fast")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape

    @pytest.mark.parametrize("shape", [(2, 10, 10), (3, 2, 10, 10)])
    def test_different_shapes(self, shape):
        s = Diffraction2D(np.ones(shape))
        s.make_probe_navigation(method="fast")
        s_nav = s._navigator_probe
        assert s.axes_manager.navigation_shape == s_nav.axes_manager.signal_shape

    @pytest.mark.parametrize("shape", [(10, 20), (2, 3, 4, 5, 6), (2, 3, 4, 5, 6, 7)])
    def test_wrong_navigation_dimensions(self, shape):
        s = Diffraction2D(np.ones(shape))
        with pytest.raises(ValueError):
            s.make_probe_navigation(method="fast")


@pytest.mark.slow
class TestPlotNavigator:
    @pytest.mark.parametrize(
        "shape", [(9, 8), (5, 9, 8), (4, 5, 9, 8), (8, 4, 5, 9, 8), (9, 8, 4, 5, 9, 8)]
    )
    def test_non_lazy(self, shape):
        s = Diffraction2D(np.random.randint(0, 256, shape), dtype=np.uint8)
        s.plot()
        plt.close("all")

    @pytest.mark.parametrize(
        "shape", [(9, 8), (5, 9, 8), (4, 5, 9, 8), (8, 4, 5, 9, 8), (9, 8, 4, 5, 9, 8)]
    )
    def test_lazy(self, shape):
        s = LazyDiffraction2D(da.random.randint(0, 256, shape), dtype=np.uint8)
        s.plot()
        plt.close("all")

    def test_navigator_kwarg(self):
        s = Diffraction2D(np.random.randint(0, 256, (8, 9, 10, 30), dtype=np.uint8))
        s_nav = Diffraction2D(np.zeros((8, 9)))
        s.plot(navigator=s_nav)
        plt.close("all")

    @pytest.mark.xfail(reason="Designed failure")
    def test_wrong_navigator_shape_kwarg(self):
        s = Diffraction2D(np.random.randint(0, 256, (8, 9, 10, 30), dtype=np.uint8))
        s_nav = Diffraction2D(np.zeros((2, 19)))
        s._navigator_probe = s_nav
        s.plot()

    def test_plot_units(self):
        s = Diffraction2D(np.random.randint(0, 256, (8, 9, 10, 30), dtype=np.uint8))
        s.axes_manager.signal_axes[0].units = "nm^-1"
        s.axes_manager.signal_axes[1].units = "nm^-1"
        s.calibration.beam_energy = 200
        s.plot(units="mrad")
        plt.close("all")


class TestFilter:
    @pytest.fixture
    def three_section(self):
        x = np.random.random((100, 50, 20, 20))
        x[0:20, :, 5:7, 5:7] = x[0:20, :, 5:7, 5:7] + 10
        x[20:60, :, 1:3, 14:16] = x[20:60, :, 1:3, 14:16] + 10
        x[60:100, :, 6:8, 10:12] = x[60:100, :, 6:8, 10:12] + 10
        d = Diffraction2D(x)
        return d

    @pytest.mark.parametrize("lazy", [True, False])
    def test_filter(self, three_section, lazy):
        if lazy:  # pragma: no cover
            dask_image = pytest.importorskip("dask_image")
            from dask_image.ndfilters import gaussian_filter as gaussian_filter

            three_section = three_section.as_lazy()
        else:
            from scipy.ndimage import gaussian_filter

        sigma = (3, 3, 3, 3)
        new = three_section.filter(func=gaussian_filter, sigma=sigma, inplace=False)
        three_section.filter(func=gaussian_filter, sigma=sigma, inplace=True)
        np.testing.assert_array_almost_equal(new.data, three_section.data)

    def test_filter_fail(self, three_section):
        def small_func(x):
            return x[1:, 1:, 1:, 1:]

        with pytest.raises(ValueError):
            new = three_section.filter(func=small_func, inplace=False)


class TestBlockwise:
    def test_blockwise(self):
        s = Diffraction2D(np.ones((2, 2, 10, 10)))
        s._blockwise(lambda x: x + 1, inplace=True)
        assert np.all(s.data == 2)
        assert s.axes_manager.signal_shape == (10, 10)

    def test_blockwise_change_signal(self):
        s = Diffraction2D(np.ones((2, 2, 10, 10)))

        def change_signal(x):
            return np.zeros((2, 2, 30, 40))

        new_s = s._blockwise(change_signal, inplace=False, signal_shape=(30, 40))
        assert np.all(new_s.data == 0)
        assert new_s.data.shape == (2, 2, 30, 40)
        assert new_s.axes_manager.signal_shape == (40, 30)

    def test_blockwise_partial_ragged(self):
        s = Diffraction2D(np.ones((2, 2, 10, 10)))

        def change_signal(x):
            exp = np.empty(
                (
                    2,
                    2,
                ),
                dtype=object,
            )
            exp[0, 0] = np.zeros((3, 4))
            exp[0, 1] = np.zeros((4, 4))
            exp[1, 0] = np.zeros((3, 4))
            exp[1, 1] = np.zeros((4, 4))
            return exp

        new_s = s._blockwise(
            change_signal, inplace=False, signal_shape=(), ragged=True, dtype=object
        )
        assert new_s.data.shape == (2, 2)
        assert new_s.axes_manager.signal_shape == ()
        assert new_s.data[0, 0].shape == (3, 4)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da

from pyxem.signals import Correlation1D, PolarDiffraction2D, LazyCorrelation1D


class TestCorrelation1D:
    @pytest.fixture
    def flat_pattern(self):
        pd = Correlation1D(data=np.ones(shape=(2, 2, 20)))
        pd.axes_manager.signal_axes[0].scale = 0.5
        pd.axes_manager.signal_axes[0].name = "theta"
        return pd

    @pytest.mark.parametrize("method", ["max", "first"])
    def test_symmetry_stem_max_first(self, flat_pattern, method):
        sym_coeff = flat_pattern.get_symmetry_coefficient(method=method)
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0])

    def test_symmetry_stem_max(self, flat_pattern):
        sym_coeff = flat_pattern.get_symmetry_coefficient(method="max")
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0])

    def test_symmetry_stem_average(self, flat_pattern):
        sym_coeff = flat_pattern.get_symmetry_coefficient(
            method="average", angular_range=0.01, normalize=True
        )
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0])

    def test_method_not_supported(self, flat_pattern):
        with pytest.raises(ValueError):
            flat_pattern.get_symmetry_coefficient(
                method="avg", angular_range=0.01, normalize=True
            )

    def test_include_duplicates(self, flat_pattern):
        sym_coeff = flat_pattern.get_symmetry_coefficient(
            method="average",
            angular_range=0.02,
            include_duplicates=True,
            normalize=True,
            symmetries=[3, 4, 5, 6, 7, 8],
        )
        np.testing.assert_array_almost_equal(sym_coeff.data, sym_coeff.data[0, 0, 0])

    def test_corr(self):
        rand_number = np.random.random(10)
        tiled_random = np.tile(rand_number, (2, 3, 2))
        p = PolarDiffraction2D(tiled_random)
        cor = p.get_pearson_correlation()
        sym_coeff = cor.get_symmetry_coefficient(
            method="average",
            angular_range=0.0,
            include_duplicates=True,
            normalize=False,
        )
        np.testing.assert_almost_equal(sym_coeff.data[0, 0], 1)

    def test_2_fold_corr(self):
        rand_number = np.random.random(10)
        tiled_random = np.tile(rand_number, (2, 3, 2))
        p = PolarDiffraction2D(tiled_random)
        cor = p.get_pearson_correlation()
        sym_coeff = cor.get_symmetry_coefficient(
            method="average",
            angular_range=0.0,
            include_duplicates=True,
            normalize=False,
        )
        np.testing.assert_almost_equal(sym_coeff.data[0, 0], 1)

    def test_lazy(self, flat_pattern):
        assert isinstance(flat_pattern, Correlation1D)
        assert isinstance(flat_pattern.as_lazy(), LazyCorrelation1D)
        assert isinstance(flat_pattern.as_lazy().data, da.Array)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


"""Dummy data for testing and examples.
"""

from .dummy_data import (
    get_disk_shift_simple_test_signal,
    get_holz_simple_test_signal,
    get_holz_heterostructure_test_signal,
    get_single_ring_diffraction_signal,
    get_dead_pixel_signal,
    get_hot_pixel_signal,
    get_simple_beam_shift_signal,
    get_stripe_pattern_beam_shift_signal,
    get_magnetic_square_beam_shift_signal,
    get_fem_signal,
    get_simple_fem_signal,
    get_generic_fem_signal,
    get_cbed_signal,
    get_simple_ellipse_signal_peak_array,
    get_nanobeam_electron_diffraction_signal,
)
from .make_diffraction_test_data import (
    EllipseRing,
    EllipseDisk,
    Circle,
    Disk,
    Ring,
    MakeTestData,
    generate_4d_data,
    DiffractionTestImage,
    DiffractionTestDataset,
)
from .make_overlapping_nanocrystals import CrystalSTEMSimulation


__all__ = [
    "get_disk_shift_simple_test_signal",
    "get_holz_simple_test_signal",
    "get_holz_heterostructure_test_signal",
    "get_single_ring_diffraction_signal",
    "get_dead_pixel_signal",
    "get_hot_pixel_signal",
    "get_simple_dpc_signal",
    "get_stripe_pattern_dpc_signal",
    "get_square_dpc_signal",
    "get_fem_signal",
    "get_simple_fem_signal",
    "get_generic_fem_signal",
    "get_cbed_signal",
    "get_simple_ellipse_signal_peak_array",
    "get_nanobeam_electron_diffraction_signal",
    "EllipseRing",
    "EllipseDisk",
    "Circle",
    "Disk",
    "Ring",
    "MakeTestData",
    "generate_4d_data",
    "DiffractionTestImage",
    "DiffractionTestDataset",
    "CrystalSTEMSimulation",
]



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for azimuthal integration."""

import numpy as np

from shapely import Polygon, box
import shapely
from numba import cuda, prange
import numba

from pyxem import CUPY_INSTALLED

if CUPY_INSTALLED:
    import cupy as cp


@numba.njit(parallel=True, nogil=True)
def _slice_radial_integrate(
    img,
    factors,
    factors_slice,
    slices,
    npt_rad,
    npt_azim,
    mask=None,
    mean=False,
):  # pragma: no cover
    """Slice the image into small chunks and multiply by the factors.

    Parameters
    ----------
    img: np.array
        The image to be sliced
    factors:
        The factors to multiply the slices by
    slices:
        The slices to slice the image by
    npt_rad:
        The number of radial points
    npt_azim:
        The number of azimuthal points

    Note
    ----
    This function is much faster with numba than without. There is probably a factor
    of 2-10 speedup that could be achieved  by using cython or c++ instead of python

    """
    if mask is not None:
        img = img * np.logical_not(mask)
    val = np.empty((npt_rad, npt_azim))
    for i in prange(len(factors_slice)):
        ii, jj = i // npt_azim, i % npt_azim
        if mean:  # divide by the total number of pixels
            total = np.sum(
                factors[factors_slice[i][0] : factors_slice[i][1]].reshape(
                    (slices[i][2] - slices[i][0], slices[i][3] - slices[i][1])
                )
            )
            if total == 0:
                val[ii, jj] = 0
            else:
                val[ii, jj] = np.sum(
                    img[slices[i][0] : slices[i][2], slices[i][1] : slices[i][3]]
                    * factors[factors_slice[i][0] : factors_slice[i][1]].reshape(
                        (slices[i][2] - slices[i][0], slices[i][3] - slices[i][1])
                    )
                ) / np.sum(
                    factors[factors_slice[i][0] : factors_slice[i][1]].reshape(
                        (slices[i][2] - slices[i][0], slices[i][3] - slices[i][1])
                    )
                )
        else:
            val[ii, jj] = np.sum(
                img[slices[i][0] : slices[i][2], slices[i][1] : slices[i][3]]
                * factors[factors_slice[i][0] : factors_slice[i][1]].reshape(
                    (slices[i][2] - slices[i][0], slices[i][3] - slices[i][1])
                )
            )
    return val


def _slice_radial_integrate_cupy(
    img, factors, factors_slice, slices, mask, npt, npt_azim
):
    original_nav = img.shape[:-2]
    img = img.reshape((-1,) + img.shape[-2:])
    val = cp.empty((img.shape[0], npt, npt_azim))
    if mask is None:
        mask = cp.zeros((img.shape[-2:]))
    __slice_radial_integrate_cupy[(img.shape[0], npt), (npt_azim)](
        img, factors, factors_slice, slices, npt_azim, mask, val
    )
    val = val.reshape(original_nav + (npt, npt_azim))
    return val


@cuda.jit
def __slice_radial_integrate_cupy(
    img, factors, factors_slice, slices, npt_azim, mask, val
):  # pragma: no cover
    """Slice the image into small chunks and multiply by the factors.
    Parameters
    ----------
    img: np.array
        The image to be sliced
    factors:
        The factors to multiply the slices by
    slices:
        The slices to slice the image by
    val:
        The array to store the result in
    Note
    ----
    This function is run by every single thread once!
    """
    tx = cuda.threadIdx.x  # current thread (azimuthal)
    bx = cuda.blockIdx.x  # Current block (navigation flattened)
    by = cuda.blockIdx.y  # Current block (radial)
    pos = cuda.grid(1)  # current thread
    index = tx + npt_azim * by
    if pos < val.size:  # account for slices out of range!
        factors_ind = factors_slice[index]
        current_slice = slices[index]
        sum = 0
        ind = 0
        for i in range(current_slice[0], current_slice[2]):
            for j in range(current_slice[1], current_slice[3]):
                is_mask = not mask[i, j]
                sum += factors[ind + factors_ind[0]] * img[bx, i, j] * is_mask
                ind += 1
        val[bx, by, tx] = sum
    return


@numba.njit
def _slice_radial_integrate1d(
    img, indexes, factors, factor_slices, mask=None, mean=False
):  # pragma: no cover
    """Slice the image into small chunks and multiply by the factors.

    Parameters
    ----------
    img: np.array
        The image to be sliced
    indexes:
        The indexes of the pixels to multiply by the `factors`
    factors:
        The percentage of the pixel for each radial bin associated with some index
    factor_slices:
        The slices to slice the factors and the indexes by
    mask:
        The mask to apply to the image
    mean:
        If True, return the mean of the pixels in the slice rather than the sum

    Note
    ----
    This function is much faster with numba than without. Additionally,  a GPU version of
    this function is not implemented because it is a bit more complicated than the 2D
    version and doesn't perform well using the `map` function.
    """
    if mask is not None:
        img = img * np.logical_not(mask)
    ans = np.empty(len(factor_slices) - 1)
    for i in range(len(factor_slices) - 1):
        ind = indexes[factor_slices[i] : factor_slices[i + 1]]
        f = factors[factor_slices[i] : factor_slices[i + 1]]
        total = 0.0
        for index, fa in zip(ind, f):
            total = total + img[index[0], index[1]] * fa
        if mean:
            total_f = np.finfo(np.float32).eps
            if mask is not None:
                for index, fa in zip(ind, f):
                    if not mask[index[0], index[1]]:
                        total_f = total_f + fa
            else:
                for index, fa in zip(ind, f):
                    total_f = total_f + fa
            total = total / total_f
        ans[i] = total
    return ans


def _get_factors(control_points, slices, pixel_extents):
    """This function takes a set of control points (vertices of bounding polygons) and
    slices (min and max indices for each control point) and returns the factors for
    each slice. The factors are the area of the intersection of the polygon and the
    sliced pixels.
    """
    all_boxes = get_boxes(slices, pixel_extent=pixel_extents)
    max_num = np.max([len(x) for x in all_boxes])
    num_box = len(all_boxes)
    boxes = shapely.empty((num_box, max_num))

    p = shapely.polygons(control_points)
    for i, bx in enumerate(all_boxes):
        try:
            b = shapely.box(bx[:, 0], bx[:, 1], bx[:, 2], bx[:, 3])
            boxes[i, : len(b)] = b
        except IndexError:  # the box is empty.
            pass

    factors = shapely.area(
        shapely.intersection(boxes, p[:, np.newaxis])
    ) / shapely.area(boxes)
    not_nan = np.logical_not(np.isnan(factors))

    factors = factors.flatten()
    factors = factors[not_nan.flatten()]

    num = np.sum(not_nan, axis=1)
    factors_slice = np.cumsum(num)
    factors_slice = np.hstack(([0], factors_slice))
    factors_slice = np.stack((factors_slice[:-1], factors_slice[1:])).T
    return factors, factors_slice


def get_boxes(slices, pixel_extent):
    all_boxes = []
    x_extent, y_extent = pixel_extent
    x_ext_left, x_ext_right = x_extent
    y_ext_left, y_ext_right = y_extent
    for sl in slices:
        x_edges = list(range(sl[0], sl[2]))
        y_edges = list(range(sl[1], sl[3]))
        boxes = []
        for i, x in enumerate(x_edges):
            for j, y in enumerate(y_edges):
                b = [
                    x_ext_left[x],
                    y_ext_left[y],
                    x_ext_right[x],
                    y_ext_right[y],
                ]
                boxes.append(b)
        all_boxes.append(np.array(boxes))
    return all_boxes


def _get_control_points(npt, npt_azim, radial_range, azimuthal_range, affine):
    """Get the control points in the form of an array (npt_azim*npt, 4, 2) representing
    the cartesian coordinates of the control points for each azimuthal pixel.

    Parameters
    ----------
    npt: int
        The number of radial points
    npt_azim:
        The number of azimuthal points
    affine: (3x3)
        The affine transformation to apply to the data
    center: (float, float)
        The center of the diffraction pattern
    radial_range: (float, float)
        The radial range of the data
    azimuthal_range: (float, float)
        The azumuthal range of the data, in radians

    Returns
    -------
    control_points: (npt_azim*npt, 4, 2)
        The cartesian coordinates of the control points of the polygon for each azimuthal pixel.

    """
    r = np.linspace(radial_range[0], radial_range[1], npt + 1)
    phi = np.linspace(azimuthal_range[0], azimuthal_range[1], npt_azim + 1)
    control_points = np.empty(((len(r) - 1) * (len(phi) - 1), 4, 2))
    # lower left
    control_points[:, 0, 0] = (np.cos(phi[:-1]) * r[:-1][:, np.newaxis]).ravel()
    control_points[:, 0, 1] = (np.sin(phi[:-1]) * r[:-1][:, np.newaxis]).ravel()
    # lower right
    control_points[:, 1, 0] = (np.cos(phi[1:]) * r[:-1][:, np.newaxis]).ravel()
    control_points[:, 1, 1] = (np.sin(phi[1:]) * r[:-1][:, np.newaxis]).ravel()
    # upper left
    control_points[:, 2, 0] = (np.cos(phi[1:]) * r[1:][:, np.newaxis]).ravel()
    control_points[:, 2, 1] = (np.sin(phi[1:]) * r[1:][:, np.newaxis]).ravel()
    # upper right
    control_points[:, 3, 0] = (np.cos(phi[:-1]) * r[1:][:, np.newaxis]).ravel()
    control_points[:, 3, 1] = (np.sin(phi[:-1]) * r[1:][:, np.newaxis]).ravel()

    # apply the affine transformation to the control points
    if affine is not None:
        affine[0, 1] = -affine[0, 1]  # changing the rotation direction
        affine[1, 0] = -affine[1, 0]
        control_points = np.dot(control_points, affine[:2, :2])
    return control_points



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from numpy.random import randint
from pyxem.signals import Diffraction2D
import pyxem.utils._cluster_tools as ct


class TestFilterPeakList:
    def test_simple(self):
        peak_list = [
            [128, 129],
            [10, 0],
            [0, 120],
            [255, 123],
            [123, 255],
            [255, 255],
            [0, 0],
        ]
        peak_list_filtered = ct._filter_peak_list(peak_list)
        assert [[128, 129]] == peak_list_filtered

    def test_max_x_index(self):
        peak_list = [
            [128, 129],
            [10, 0],
            [0, 120],
            [256, 123],
            [123, 256],
            [256, 256],
            [0, 0],
        ]
        peak_list_filtered = ct._filter_peak_list(peak_list, max_x_index=256)
        assert [[128, 129], [123, 256]] == peak_list_filtered

    def test_max_y_index(self):
        peak_list = [
            [128, 129],
            [10, 0],
            [0, 120],
            [256, 123],
            [123, 256],
            [256, 256],
            [0, 0],
        ]
        peak_list_filtered = ct._filter_peak_list(peak_list, max_y_index=256)
        assert [[128, 129], [256, 123]] == peak_list_filtered


class TestFilterPeakListRadius:
    def test_simple(self):
        peak_list = np.random.randint(100, size=(1000, 2))
        peak_filtered_list0 = ct._filter_peak_list_radius(
            peak_list, xc=50, yc=50, r_min=30
        )
        assert len(peak_list) != len(peak_filtered_list0)
        peak_filtered_list1 = ct._filter_peak_list_radius(
            peak_list, xc=50, yc=50, r_min=1000
        )
        assert len(peak_filtered_list1) == 0

    def test_r_lim(self):
        peak_list = np.array([[50, 50], [50, 30]])
        peak_filtered_list0 = ct._filter_peak_list_radius(
            peak_list, xc=50, yc=50, r_min=19
        )
        assert (peak_filtered_list0 == np.array([[50, 30]])).all()
        peak_filtered_list1 = ct._filter_peak_list_radius(
            peak_list, xc=50, yc=50, r_min=21
        )
        assert len(peak_filtered_list1) == 0

    def test_r_max(self):
        peak_list = np.array([[50, 0], [0, 30]])
        peak_filtered_list0 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_max=40
        )
        assert (peak_filtered_list0 == np.array([[0, 30]])).all()
        peak_filtered_list1 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_max=52
        )
        assert (peak_filtered_list1 == np.array([[50, 0], [0, 30]])).all()
        peak_filtered_list2 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_max=81
        )
        assert len(peak_filtered_list2) == 2
        peak_filtered_list3 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_max=11
        )
        assert len(peak_filtered_list3) == 0

    def test_r_min_and_r_max(self):
        peak_list = np.array([[50, 0], [0, 30]])
        peak_filtered_list0 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_min=20, r_max=40
        )
        assert (peak_filtered_list0 == np.array([[0, 30]])).all()
        peak_filtered_list1 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_min=20, r_max=60
        )
        assert (peak_filtered_list1 == np.array([[50, 0], [0, 30]])).all()
        peak_filtered_list2 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_min=40, r_max=60
        )
        assert (peak_filtered_list2 == np.array([[50, 0]])).all()
        peak_filtered_list3 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_min=10, r_max=20
        )
        assert len(peak_filtered_list3) == 0
        peak_filtered_list4 = ct._filter_peak_list_radius(
            peak_list, xc=0, yc=0, r_min=90, r_max=99
        )
        assert len(peak_filtered_list4) == 0

    def test_xc_yc(self):
        peak_list = np.array([[50, 10], [50, 50]])
        peak_filtered_list = ct._filter_peak_list_radius(
            peak_list, xc=10, yc=50, r_min=10
        )
        assert (peak_filtered_list == np.array([[50, 50]])).all()

    def test_wrong_input_no_r_min_or_r_max(self):
        peak_list = np.array([[50, 10], [50, 50]])
        with pytest.raises(ValueError):
            ct._filter_peak_list_radius(peak_list, xc=10, yc=50)

    def test_r_min_larger_than_r_max(self):
        peak_list = np.array([[50, 10], [50, 50]])
        with pytest.raises(ValueError):
            ct._filter_peak_list_radius(peak_list, xc=10, yc=50, r_min=50, r_max=30)


class TestFilterPeakArrayRadius:
    def test_simple(self):
        peak_array = np.empty(shape=(2, 3), dtype=object)
        for ix, iy in np.ndindex(peak_array.shape):
            peak_array[ix, iy] = np.random.randint(30, 70, size=(1000, 2))
        peak_array_filtered = ct._filter_peak_array_radius(peak_array, 50, 50, r_min=10)
        assert peak_array_filtered.shape == (2, 3)
        for ix, iy in np.ndindex(peak_array_filtered.shape):
            assert len(peak_array_filtered[ix, iy]) != 1000

    def test_r_lim(self):
        peak_array = np.empty(shape=(2, 3), dtype=object)
        for ix, iy in np.ndindex(peak_array.shape):
            peak_array[ix, iy] = np.random.randint(30, 70, size=(1000, 2))
        peak_array_filtered = ct._filter_peak_array_radius(peak_array, 50, 50, r_min=50)
        for ix, iy in np.ndindex(peak_array_filtered.shape):
            assert len(peak_array_filtered[ix, iy]) == 0

    def test_r_max(self):
        peak_array = np.empty(shape=(2, 3), dtype=object)
        for ix, iy in np.ndindex(peak_array.shape):
            peak_array[ix, iy] = np.random.randint(30, 70, size=(1000, 2))
        peak_array_filtered0 = ct._filter_peak_array_radius(peak_array, 0, 0, r_max=20)
        for ix, iy in np.ndindex(peak_array_filtered0.shape):
            assert len(peak_array_filtered0[ix, iy]) == 0
        peak_array_filtered1 = ct._filter_peak_array_radius(peak_array, 0, 0, r_max=100)
        for ix, iy in np.ndindex(peak_array_filtered1.shape):
            assert len(peak_array_filtered1[ix, iy]) == 1000

    def test_r_min_and_r_max(self):
        peak_array = np.empty(shape=(2, 3), dtype=object)
        for ix, iy in np.ndindex(peak_array.shape):
            peak_array[ix, iy] = np.random.randint(30, 70, size=(1000, 2))
        peak_array_filtered0 = ct._filter_peak_array_radius(
            peak_array, 0, 0, r_min=20, r_max=25
        )
        for ix, iy in np.ndindex(peak_array_filtered0.shape):
            assert len(peak_array_filtered0[ix, iy]) == 0
        peak_array_filtered1 = ct._filter_peak_array_radius(
            peak_array, 0, 0, r_min=20, r_max=100
        )
        for ix, iy in np.ndindex(peak_array_filtered1.shape):
            assert len(peak_array_filtered1[ix, iy]) == 1000
        peak_array_filtered2 = ct._filter_peak_array_radius(
            peak_array, 0, 0, r_min=110, r_max=130
        )
        for ix, iy in np.ndindex(peak_array_filtered2.shape):
            assert len(peak_array_filtered2[ix, iy]) == 0

    def test_xc(self):
        peak_array = np.empty(shape=(2, 3), dtype=object)
        for ix, iy in np.ndindex(peak_array.shape):
            peak_array[ix, iy] = np.random.randint(30, 70, size=(1000, 2))
        peak_array_filtered = ct._filter_peak_array_radius(peak_array, 10, 50, r_min=10)
        for ix, iy in np.ndindex(peak_array_filtered.shape):
            assert len(peak_array_filtered[ix, iy]) == 1000

    def test_yc(self):
        peak_array = np.empty(shape=(2, 3), dtype=object)
        for ix, iy in np.ndindex(peak_array.shape):
            peak_list = np.random.randint(30, 70, size=(999, 2)).tolist()
            peak_list.append([10, 50])
            peak_array[ix, iy] = np.array(peak_list)
        peak_array_filtered = ct._filter_peak_array_radius(peak_array, 50, 10, r_min=10)
        for ix, iy in np.ndindex(peak_array_filtered.shape):
            assert len(peak_array_filtered[ix, iy]) == 999


class TestFindMaxIndices4DPeakArray:
    def test_simple(self):
        max_x, max_y = 255, 127
        peak_array0 = np.random.randint(10, max_x, size=(3, 4, 5000, 1))
        peak_array1 = np.random.randint(5, max_y, size=(3, 4, 5000, 1))
        peak_array = np.concatenate((peak_array0, peak_array1), axis=3)
        max_x_index, max_y_index = ct._find_max_indices_4D_peak_array(peak_array)
        assert max_x_index == max_x - 1
        assert max_y_index == max_y - 1


class TestFilter4DPeakArray:
    def test_simple(self):
        peak_array0 = randint(124, 132, size=(3, 4, 10, 2))
        peak_array1 = np.ones(shape=(3, 4, 3, 2)) * 255
        peak_array2 = np.zeros(shape=(3, 4, 3, 2))
        peak_array = np.concatenate((peak_array0, peak_array1, peak_array2), axis=2)
        peak_array_filtered = ct._filter_4D_peak_array(peak_array)
        for ix, iy in np.ndindex(peak_array_filtered.shape[:2]):
            peak_list = peak_array_filtered[ix, iy]
            for x, y in peak_list:
                assert x != 0
                assert x != 255
                assert y != 0
                assert y != 255

    def test_max_x_index_max_y_index(self):
        peak_array0 = randint(124, 132, size=(3, 4, 10, 2))
        peak_array1 = np.ones(shape=(3, 4, 3, 2)) * 256
        peak_array2 = np.zeros(shape=(3, 4, 3, 2))
        peak_array = np.concatenate((peak_array0, peak_array1, peak_array2), axis=2)
        peak_array_filtered = ct._filter_4D_peak_array(
            peak_array, max_x_index=256, max_y_index=256
        )
        for ix, iy in np.ndindex(peak_array_filtered.shape[:2]):
            peak_list = peak_array_filtered[ix, iy]
            for x, y in peak_list:
                assert x != 0
                assert x != 256
                assert y != 0
                assert y != 256

    def test_signal_axes(self):
        s = Diffraction2D(np.zeros(shape=(3, 4, 128, 128)))
        peak_array0 = randint(62, 67, size=(3, 4, 10, 2))
        peak_array1 = np.ones(shape=(3, 4, 3, 2)) * 127
        peak_array2 = np.zeros(shape=(3, 4, 3, 2))
        peak_array = np.concatenate((peak_array0, peak_array1, peak_array2), axis=2)
        peak_array_filtered = ct._filter_4D_peak_array(
            peak_array, signal_axes=s.axes_manager.signal_axes
        )
        for ix, iy in np.ndindex(peak_array_filtered.shape[:2]):
            peak_list = peak_array_filtered[ix, iy]
            for x, y in peak_list:
                assert x != 0
                assert x != 127
                assert y != 0
                assert y != 127

    def test_1d_nav(self):
        peak_array = randint(124, 132, size=(4, 10, 2))
        peak_array_filtered = ct._filter_4D_peak_array(peak_array)
        assert peak_array_filtered.shape == (4,)

    def test_3d_nav(self):
        peak_array = randint(124, 132, size=(2, 3, 4, 10, 2))
        peak_array_filtered = ct._filter_4D_peak_array(peak_array)
        assert peak_array_filtered.shape == (2, 3, 4)


class TestGetClusterDict:
    def test_simple(self):
        peak_array = randint(100, size=(100, 2))
        ct._get_cluster_dict(peak_array)

    def test_eps(self):
        peak_array0 = randint(6, size=(100, 2)) + 80
        peak_array1 = randint(6, size=(100, 2))
        peak_array = np.vstack((peak_array0, peak_array1))
        cluster_dict0 = ct._get_cluster_dict(peak_array)
        assert len(cluster_dict0) == 2
        assert len(cluster_dict0[0]) == 100
        assert len(cluster_dict0[1]) == 100
        cluster_dict1 = ct._get_cluster_dict(peak_array, eps=200)
        assert len(cluster_dict1) == 1
        assert len(cluster_dict1[0]) == 200

    def test_min_samples(self):
        peak_array0 = randint(6, size=(100, 2))
        peak_array = np.vstack((peak_array0, [[54, 21], [53, 20], [55, 22]]))
        cluster_dict0 = ct._get_cluster_dict(peak_array, min_samples=2)
        labels0 = sorted(list(cluster_dict0.keys()))
        assert labels0 == [0, 1]
        cluster_dict1 = ct._get_cluster_dict(peak_array, min_samples=4)
        labels1 = sorted(list(cluster_dict1.keys()))
        assert labels1 == [-1, 0]

    def test_three_clusters(self):
        peak_array0 = randint(6, size=(100, 2)) + 80
        peak_array1 = randint(6, size=(100, 2))
        peak_array = np.vstack(
            (
                peak_array0,
                peak_array1,
                [
                    [54, 21],
                ],
            )
        )
        cluster_dict = ct._get_cluster_dict(peak_array, min_samples=2)
        labels = sorted(list(cluster_dict.keys()))
        assert labels == [-1, 0, 1]


class TestSortClusterDict:
    def test_simple(self):
        n_centre, n_rest = 10, 20
        cluster_dict = {}
        cluster_dict[-1] = [
            [5, 100],
        ]
        cluster_dict[0] = randint(5, size=(n_centre, 2)).tolist()
        cluster_dict[1] = randint(100, 105, size=(n_rest, 2)).tolist()
        sorted_cluster_dict0 = ct._sort_cluster_dict(
            cluster_dict, centre_x=2, centre_y=2
        )
        assert len(sorted_cluster_dict0["centre"]) == n_centre
        assert len(sorted_cluster_dict0["rest"]) == n_rest
        assert len(sorted_cluster_dict0["none"]) == 1

        sorted_cluster_dict1 = ct._sort_cluster_dict(
            cluster_dict, centre_x=102, centre_y=102
        )
        assert len(sorted_cluster_dict1["centre"]) == n_rest
        assert len(sorted_cluster_dict1["rest"]) == n_centre
        assert len(sorted_cluster_dict1["none"]) == 1


class TestClusterAndSortPeakArray:
    def test_simple(self):
        peak_array0 = randint(124, 132, size=(3, 4, 10, 2))
        peak_array1 = randint(24, 32, size=(3, 4, 5, 2))
        peak_array2 = randint(201, 203, size=(3, 4, 1, 2))
        peak_array = np.concatenate((peak_array0, peak_array1, peak_array2), axis=2)
        peak_dict = ct._cluster_and_sort_peak_array(peak_array)
        assert len(peak_dict["centre"][0, 0]) == 10
        assert len(peak_dict["rest"][0, 0]) == 5
        assert len(peak_dict["none"][0, 0]) == 1

    def test_only_centre(self):
        peak_array = randint(124, 132, size=(3, 4, 10, 2))
        peak_dict = ct._cluster_and_sort_peak_array(peak_array)
        assert len(peak_dict["centre"][0, 0]) == 10
        assert len(peak_dict["rest"][0, 0]) == 0
        assert len(peak_dict["none"][0, 0]) == 0

    def test_only_none(self):
        peak_array = randint(201, 203, size=(3, 4, 1, 2))
        peak_dict = ct._cluster_and_sort_peak_array(peak_array)
        assert len(peak_dict["centre"][0, 0]) == 0
        assert len(peak_dict["rest"][0, 0]) == 0
        assert len(peak_dict["none"][0, 0]) == 1

    def test_eps(self):
        peak_array0 = randint(124, 132, size=(3, 4, 10, 2))
        peak_array1 = randint(24, 32, size=(3, 4, 5, 2))
        peak_array = np.concatenate((peak_array0, peak_array1), axis=2)
        peak_dict0 = ct._cluster_and_sort_peak_array(peak_array, eps=30)
        peak_dict1 = ct._cluster_and_sort_peak_array(peak_array, eps=300)
        assert len(peak_dict0["centre"][0, 0]) == 10
        assert len(peak_dict1["centre"][0, 0]) == 15

    def test_min_samples(self):
        peak_array0 = randint(124, 132, size=(3, 4, 10, 2))
        peak_array1 = randint(204, 208, size=(3, 4, 3, 2))
        peak_array = np.concatenate((peak_array0, peak_array1), axis=2)
        peak_dict0 = ct._cluster_and_sort_peak_array(peak_array, min_samples=4)
        peak_dict1 = ct._cluster_and_sort_peak_array(peak_array, min_samples=2)
        assert len(peak_dict0["none"][0, 0]) == 3
        assert len(peak_dict1["none"][0, 0]) == 0
        assert len(peak_dict1["rest"][0, 0]) == 3

    def test_different_centre(self):
        peak_array0 = randint(124, 132, size=(3, 4, 10, 2))
        peak_array1 = randint(24, 32, size=(3, 4, 5, 2))
        peak_array = np.concatenate((peak_array0, peak_array1), axis=2)
        peak_dict0 = ct._cluster_and_sort_peak_array(
            peak_array, centre_x=128, centre_y=128
        )
        peak_dict1 = ct._cluster_and_sort_peak_array(
            peak_array, centre_x=28, centre_y=28
        )
        assert len(peak_dict0["centre"][0, 0]) == 10
        assert len(peak_dict0["rest"][0, 0]) == 5
        assert len(peak_dict1["centre"][0, 0]) == 5
        assert len(peak_dict1["rest"][0, 0]) == 10

    @pytest.mark.parametrize("ndim", [0, 1, 2, 3, 4, 5, 6])
    def test_ndim(self, ndim):
        nav_shape = tuple(np.random.randint(2, 5, size=ndim))
        shape = nav_shape + (23, 2)
        peak_array = np.random.randint(99, size=shape)
        peak_dict = ct._cluster_and_sort_peak_array(peak_array)
        assert peak_dict["centre"].shape == nav_shape
        assert peak_dict["rest"].shape == nav_shape
        assert peak_dict["none"].shape == nav_shape


class TestAddPeakDictsToSignal:
    def test_simple(self):
        peak_dicts = {}
        center_arr = np.empty(shape=(4, 3), dtype=object)
        rest_arr = np.empty(shape=(4, 3), dtype=object)
        none_arr = np.empty(shape=(4, 3), dtype=object)
        for ind in np.ndindex(center_arr.shape):
            center_arr[ind] = np.random.randint(124, 132, size=(10, 2))
            rest_arr[ind] = np.random.randint(204, 212, size=(5, 2))
            none_arr[ind] = np.random.randint(10, 13, size=(2, 2))
        peak_dicts["centre"] = center_arr
        peak_dicts["rest"] = rest_arr
        peak_dicts["none"] = none_arr
        s = Diffraction2D(np.zeros((3, 4, 256, 256)))
        ct._add_peak_dicts_to_signal(s, peak_dicts)


class TestGetPeakArrayShape:
    @pytest.mark.parametrize("ndim", [0, 1, 2, 3, 4, 5, 6])
    def test_dtype_object(self, ndim):
        shape = tuple(np.random.randint(2, 5, size=ndim))
        peak_array = np.empty(shape, dtype=object)
        shape_output = ct._get_peak_array_shape(peak_array)
        assert shape == shape_output

    @pytest.mark.parametrize("ndim", [0, 1, 2, 3, 4, 5, 6])
    def test_not_dtype_object(self, ndim):
        nav_shape = tuple(np.random.randint(2, 5, size=ndim))
        shape = nav_shape + (23, 2)
        peak_array = np.random.randint(99, size=shape)
        shape_output = ct._get_peak_array_shape(peak_array)
        assert nav_shape == shape_output


class TestSortedClusterDictToMarkerList:
    def test_simple(self):
        sorted_cluster_dict = {}
        sorted_cluster_dict["centre"] = randint(10, size=(3, 4, 2, 2))
        sorted_cluster_dict["rest"] = randint(50, 60, size=(3, 4, 3, 2))
        sorted_cluster_dict["none"] = randint(90, size=(3, 4, 1, 2))
        marker_list = ct._sorted_cluster_dict_to_marker_list(sorted_cluster_dict)
        assert len(marker_list) == 3

    def test_size(self):
        marker_size = 30
        sorted_cluster_dict = {}
        sorted_cluster_dict["centre"] = randint(10, size=(3, 4, 2, 2))
        sorted_cluster_dict["rest"] = randint(50, 60, size=(3, 4, 3, 2))
        sorted_cluster_dict["none"] = randint(90, size=(3, 4, 1, 2))
        marker_list = ct._sorted_cluster_dict_to_marker_list(
            sorted_cluster_dict, size=marker_size
        )

    def test_color(self):
        marker_color = "orange"
        sorted_cluster_dict = {}
        sorted_cluster_dict["centre"] = randint(10, size=(3, 4, 2, 2))
        sorted_cluster_dict["rest"] = randint(50, 60, size=(3, 4, 3, 2))
        sorted_cluster_dict["none"] = randint(90, size=(3, 4, 1, 2))
        sorted_cluster_dict["magic"] = randint(40, size=(3, 4, 1, 2))
        marker_list = ct._sorted_cluster_dict_to_marker_list(
            sorted_cluster_dict,
            color_rest=marker_color,
            color_centre=marker_color,
            color_none=marker_color,
        )
